import sys
import os
import pkgutil
import argparse
import importlib


class Baseporter:
    def __init__(self, name, extensions, callback, extra_options=[], generic_options=set(), slug=None):
        self.name = name
        self.extensions = extensions
        self.callback = callback
        self.extra_options = extra_options
        self.generic_options = generic_options
        self.slug = slug if slug is not None else extensions[0]

    def process(self, *a, **kw):
        return self.callback(*a, **kw)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.slug)

    def argparse_options(self, ns):
        o_options = {}
        for opt in self.extra_options:
            o_options[opt.dest] = getattr(ns, opt.nsvar(self.slug))
        for opt in self.generic_options:
            o_options[opt] = getattr(ns, opt)
        return o_options


class ExtraOption:
    def __init__(self, name, **kwargs):
        self.name = name
        self.kwargs = kwargs
        if "action" not in self.kwargs:
            self.kwargs["metavar"] = self.name
        self.dest = kwargs.pop("dest", name)

    def add_argument(self, slug, parser):
        opt = "--%s-%s" % (slug, self.name.replace("_", "-"))
        parser.add_argument(opt, dest=self.nsvar(slug), **self.kwargs)

    def nsvar(self, slug):
        return "%s_%s" % (slug, self.dest)


def _add_options(parser, ie, object):
    if not object.extra_options:
        return

    suf = " %sing options" % ie
    group = parser.add_argument_group(object.name + suf)
    for op in object.extra_options:
        op.add_argument(object.slug, group)


class Loader:
    def __init__(self, module_path, module_name, ie):
        self._loaded = False
        self._registry = {}
        self._module_path = os.path.dirname(module_path)
        self._module_name = module_name.replace(".base", "")
        self._ie = ie
        self._failed = {}

    def load_modules(self):
        self._loaded = True

        for _, modname, _ in pkgutil.iter_modules([self._module_path]):
            if modname == "base":
                continue

            full_modname = "." + modname
            try:
                importlib.import_module(full_modname, self._module_name)
            except ImportError as e:
                self._failed[modname] = "cairosvg or glaxnimate" if e.name == "lottie.exporters.cairo" else e.name

    @property
    def failed_modules(self):
        if not self._loaded:
            self.load_modules()

        return self._failed

    @property
    def items(self):
        if not self._loaded:
            self.load_modules()
        return self._registry

    def __iter__(self):
        return iter(self.items.values())

    def get(self, slug):
        return self.items.get(slug, None)

    def __getitem__(self, key):
        return self.get(key)

    def get_from_filename(self, filename):
        return self.get_from_extension(os.path.splitext(filename)[1][1:])

    def get_from_extension(self, ext):
        for p in self.items.values():
            if ext in p.extensions:
                return p
        return None

    def set_options(self, parser):
        for exporter in self.items.values():
            _add_options(parser, self._ie, exporter)

    def keys(self):
        return self.items.keys()

    def decorator(self, name, extensions, extra_options=[], generic_options=set(), slug=None):
        def decorator(callback):
            porter = Baseporter(name, extensions, callback, extra_options, generic_options, slug)
            self._registry[porter.slug] = porter
            return callback
        return decorator


class IoProgressReporter:
    def report_progress(self, title, value, total):
        sys.stderr.write("\r%s %s/%s" % (title, value, total))
        sys.stderr.flush()

    def report_message(self, message):
        sys.stderr.write("\r" + message + "\n")
        sys.stderr.flush()


IoProgressReporter.instance = IoProgressReporter()


def io_progress():
    return IoProgressReporter.instance



from PyQt5.QtGui import QIcon
from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtCore import pyqtSignal as Signal
from PyQt5.QtCore import pyqtSlot as Slot
from PyQt5.QtWidgets import *


class SearchWidget(QWidget):
    frame_changed = Signal(int)

    def __init__(self, target):
        super().__init__()
        self.target = target

        layout_slider = QHBoxLayout()
        layout_slider.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout_slider)

        self.line_find = QLineEdit()
        self.line_find.setPlaceholderText("Find")
        self.line_find.returnPressed.connect(self.find_next)
        layout_slider.addWidget(self.line_find)

        self.line_replace = QLineEdit()
        self.line_replace.setPlaceholderText("Replace")
        self.line_replace.returnPressed.connect(self.replace)
        layout_slider.addWidget(self.line_replace)

        self.button_next = QPushButton()
        self.button_next.setIcon(QIcon.fromTheme("go-down-search"))
        self.button_next.clicked.connect(self.find_next)
        self.button_next.setToolTip("Next")
        layout_slider.addWidget(self.button_next)

        self.button_prev = QPushButton()
        self.button_prev.setIcon(QIcon.fromTheme("go-up-search"))
        self.button_prev.clicked.connect(self.find_prev)
        self.button_next.setToolTip("Previous")
        layout_slider.addWidget(self.button_prev)

        self.button_repl = QPushButton()
        self.button_repl.setText("Replace")
        self.button_repl.clicked.connect(self.replace)
        layout_slider.addWidget(self.button_repl)

        self.button_case = QPushButton()
        self.button_case.setCheckable(True)
        self.button_case.setIcon(QIcon.fromTheme("format-text-superscript"))
        self.button_case.setToolTip("Case Sensitive")
        layout_slider.addWidget(self.button_case)

        self.combo_mode = QComboBox()
        self.combo_mode.addItem("Text")
        self.combo_mode.addItem("RegExp")
        layout_slider.addWidget(self.combo_mode)

        self.button_selection = QPushButton()
        self.button_selection.setCheckable(True)
        self.button_selection.setIcon(QIcon.fromTheme("edit-select-all"))
        self.button_selection.setToolTip("Only in selection")
        layout_slider.addWidget(self.button_selection)

        self.set_replace_enabled(False)

    def start_search(self, replace):
        if self.target.hasSelectedText():
            st = self.target.selectedText()
            if "\n" in st:
                self.button_selection.setChecked(True)
            else:
                self.button_selection.setChecked(False)
                self.line_find.setText(st)
        self.line_find.selectAll()
        self.line_find.setFocus(True)
        self.set_replace_enabled(replace)

    def set_replace_enabled(self, show):
        #self.button_selection.setChecked(show != self.button_selection.isVisible())
        #self.button_selection.setVisible(show)
        self.line_replace.setVisible(show)
        self.button_repl.setVisible(show)

    def find(self, forward):
        func = self.target.findFirst if not self.button_selection.isChecked() else self.target.findFirstInSelection
        func(
            self.line_find.text(),
            self.combo_mode.currentIndex() != 0,
            self.button_case.isChecked(),
            False,
            True,
            forward,
            True
        )

    def find_next(self):
        return self.find(True)

    def find_prev(self):
        return self.find(False)

    def replace(self):
        self.target.replace(self.line_replace.text())
        self.find_next()



from .base import LottieObject, LottieProp, PseudoBool, Index
from .layers import Layer, PreCompLayer
from .assets import Asset, Precomp
from .text import FontList, Chars
from .composition import Composition
from .helpers import VisualObject, Marker

##\defgroup Lottie Lottie
#
# Objects of the lottie file structure.

## \defgroup LottieCheck Lottie (to check)
#
# Lottie objects that have not been tested


## @ingroup Lottie
class Metadata(LottieObject):
    """!
    Document metadata
    """
    _props = [
        LottieProp("author", "a", str, False),
        LottieProp("generator", "g", str, False),
        LottieProp("keywords", "k", str, True),
        LottieProp("description", "d", str, False),
        LottieProp("theme_color", "tc", str, False),
    ]

    def __init__(self):
        self.generator = None
        self.author = None
        self.keywords = None
        self.description = None
        self.theme_color = None


#ingroup Lottie
class UserMetadata(LottieObject):
    """!
    User-defined metadata
    """
    _props = [
        LottieProp("filename", "filename", str, False),
        LottieProp("custom_properties", "customProps", dict, False),
    ]

    def __init__(self):
        super().__init__()

        self.filename = None
        self.custom_properties = {}


#ingroup Lottie
class MotionBlur(LottieObject):
    """!
    Motion blur settings
    """
    _props = [
        LottieProp("shutter_angle", "sa", float, False),
        LottieProp("shutter_phase", "sp", float, False),
        LottieProp("samples_per_frame", "spf", float, False),
        LottieProp("adaptive_sample_limit", "asl", float, False),
    ]

    def __init__(self):
        super().__init__()

        ## Angle in degrees
        self.shutter_angle = None
        ## Angle in degrees
        self.shutter_phase = None
        self.samples_per_frame = None
        self.adaptive_sample_limit = None


## @ingroup Lottie
class Animation(Composition, VisualObject):
    """!
    Top level object, describing the animation

    @see http://docs.aenhancers.com/items/compitem/
    """
    _props = [
        LottieProp("version", "v", str, False),
        LottieProp("frame_rate", "fr", float, False),
        LottieProp("in_point", "ip", float, False),
        LottieProp("out_point", "op", float, False),
        LottieProp("width", "w", int, False),
        LottieProp("height", "h", int, False),
        LottieProp("threedimensional", "ddd", PseudoBool, False),
        LottieProp("assets", "assets", Asset, True),
        LottieProp("extra_compositions", "comps", Precomp, True),
        LottieProp("fonts", "fonts", FontList),
        LottieProp("chars", "chars", Chars, True),
        LottieProp("markers", "markers", Marker, True),
        LottieProp("motion_blur", "mb", MotionBlur, False),
        LottieProp("metadata", "meta", Metadata, False),
        LottieProp("user_metadata", "metadata", UserMetadata, False),
    ]
    _version = "5.5.2"

    def __init__(self, n_frames=60, framerate=60):
        super().__init__()
        ## The time when the composition work area begins, in frames.
        self.in_point = 0
        ## The time when the composition work area ends.
        ## Sets the final Frame of the animation
        self.out_point = n_frames
        ## Frames per second
        self.frame_rate = framerate
        ## Composition Width
        self.width = 512
        ## Composition has 3-D layers
        self.threedimensional = False
        ## Composition Height
        self.height = 512
        ## Bodymovin Version
        self.version = self._version
        ## source items that can be used in multiple places. Comps and Images for now.
        self.assets = [] # Image, Precomp
        ## List of Extra compositions not referenced by anything
        self.extra_compositions = None
        ## source chars for text layers
        self.chars = None
        ## Available fonts
        self.fonts = None
        self.metadata = None
        self.user_metadata = None
        self.motion_blur = None
        self.markers = None

    def precomp(self, name):
        for ass in self.assets:
            if isinstance(ass, Precomp) and ass.id == name:
                return ass
        return None

    def _on_prepare_layer(self, layer):
        if layer.in_point is None:
            layer.in_point = self.in_point
        if layer.out_point is None:
            layer.out_point = self.out_point

    def to_precomp(self):
        """!
        Turns the main comp into a precomp
        """
        precomp = Precomp()
        #precomp.frame_rate = self.frame_rate
        precomp.layers = self.layers
        precomp.name = self.name
        name_id = 0
        base_name = self.name or "Animation"
        name = base_name
        index = 0
        while True:
            if index >= len(self.assets):
                break

            while self.assets[index].id == name:
                name_id += 1
                name = "%s %s" % (base_name, name_id)
                index = -1

            index += 1
        precomp.id = name
        self.assets.append(precomp)

        precomp_layer = PreCompLayer()
        precomp_layer.width = self.width
        precomp_layer.height = self.height
        precomp_layer.in_point = self.in_point
        precomp_layer.out_point = self.out_point
        precomp_layer.reference_id = name
        self.layers = [precomp_layer]
        return precomp

    def scale(self, width, height):
        """!
        Scales the animation so it fits in width/height
        """
        if self.width != width or self.height != height:
            self.to_precomp()

            scale = min(width/self.width, height/self.height)
            self.width = width
            self.height = height

            self.layers[0].transform.scale.value *= scale

    def tgs_sanitize(self):
        """!
        Cleans up some things to ensure it works as a telegram sticker
        """
        self.scale(512, 512)

        if self.frame_rate < 45:
            self.frame_rate = 30
        else:
            self.frame_rate = 60

    def _fixup(self):
        super()._fixup()
        if self.assets:
            for ass in self.assets:
                if isinstance(ass, Precomp):
                    ass.animation = self
                    ass._fixup()

    def __str__(self):
        return self.name or super().__str__()



from xml.dom import minidom
from xml.etree import ElementTree

from .base import exporter
from ..parsers.baseporter import ExtraOption
from ..parsers.svg.builder import to_svg
from ..utils.file import open_file


def _print_ugly_xml(dom, file):
    return dom.write(file, "utf-8", True)


def _print_pretty_xml(dom, file):
    with open_file(file) as fp:
        xmlstr = minidom.parseString(ElementTree.tostring(dom.getroot())).toprettyxml(indent="   ")
        fp.write(xmlstr)


@exporter("SVG", ["svg"], [
    ExtraOption("animated", action="store_true", help="Export animated svg"),
], {"pretty", "frame"})
def export_svg(animation, file, frame=0, pretty=True, animated=False):
    _print_xml = _print_pretty_xml if pretty else _print_ugly_xml
    _print_xml(to_svg(animation, frame, animated), file)



import zipfile
import warnings
from xml.etree import ElementTree

from .base import importer
from ..parsers.svg.importer import SvgParser
from .. import objects

ns = "{%s}" % "http://www.calligra.org/DTD/krita"


def _ns(string):
    return string.format(ns=ns)


def _import_layers(zf, animation, xml_parent, svg_parser, parent):
    for xml_layer in xml_parent.findall(_ns("./{ns}layers/{ns}layer")):
        nodetype = xml_layer.attrib["nodetype"]

        if nodetype == "grouplayer":
            layer = animation.add_layer(objects.NullLayer())
            _import_layers(zf, animation, xml_layer, svg_parser, layer)
        elif nodetype == "shapelayer":
            filename = "%s/layers/%s.shapelayer/content.svg" % (animation.name, xml_layer.attrib["filename"])
            with zf.open(filename) as svg_tree:
                layer = svg_parser.etree_to_layer(animation, ElementTree.parse(svg_tree))
        else:
            warnings.warn("Unsupported krita layer %s" % nodetype)
            continue

        layer.name = xml_layer.attrib["name"]
        if xml_layer.attrib["visible"] == 0:
            layer.transform.opacity.value = 0
        layer.parent = parent


@importer("Krita", ["kra"])
def import_krita(file):
    with zipfile.ZipFile(file) as zf:
        with zf.open("maindoc.xml") as main:
            main_xml = ElementTree.parse(main)

        image = main_xml.find(_ns("./{ns}IMAGE"))
        fps = float(main_xml.find(_ns("./{ns}IMAGE/{ns}animation/{ns}framerate")).attrib["value"])
        framerange = main_xml.find(_ns("./{ns}IMAGE/{ns}animation/{ns}range")).attrib

        animation = objects.Animation(int(framerange["to"]), fps)
        animation.in_point = int(framerange["from"])
        animation.width = int(image.attrib["width"])
        animation.height = int(image.attrib["height"])
        animation.name = image.attrib["name"]

        parser = SvgParser()
        parser.dpi = int(image.attrib["x-res"])

        _import_layers(zf, animation, image, parser, None)

    return animation



from xml.dom import minidom
import copy
import enum

from .core_nodes import XmlDescriptor, XmlSimpleElement, ValueReference
from .utils import *
from lottie.nvector import NVector
from lottie.parsers.sif.ast_impl.base import SifAstNode, SifValue
from lottie.parsers.sif.sif.core import ObjectRegistry, TypeDescriptor, noop


class XmlAnimatable(XmlDescriptor):
    def __init__(self, name, typename, default=None, type_wrapper=noop):
        super().__init__(name)
        self.type = TypeDescriptor(typename, default, type_wrapper)

    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry, param: TypeDescriptor = None):
        cn = xml_first_element_child(parent, self.name)
        if cn:
            value = SifAstNode.from_dom(xml_first_element_child(cn), self.type_for(param), registry)
        else:
            value = self.default()

        setattr(obj, self.att_name, value)

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document, type: TypeDescriptor = None):
        value = getattr(obj, self.att_name)
        if isinstance(value, SifValue) and isinstance(value.value, ValueReference):
            parent.setAttribute(self.name, ":" + value.value.id)
            return
        param = parent.appendChild(dom.createElement(self.name))
        param.appendChild(value.to_dom(dom, self.type_for(type)))
        return param

    def from_python(self, value):
        if not isinstance(value, SifAstNode):
            raise ValueError("%s isn't a valid value for %s" % (value, self.name))
        return value

    def default(self):
        return SifValue(copy.deepcopy(self.type.default_value))

    def type_for(self, param: TypeDescriptor):
        if param is not None and self.type.typename == "_recurse":
            return param
        return self.type


class XmlParam(XmlDescriptor):
    def __init__(self, name, typename, default=None, type_wrapper=noop, static=False):
        super().__init__(name)
        self.type = TypeDescriptor(typename, default, type_wrapper)
        self.static = static

    def _def(self):
        from ..api import Def
        return Def

    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        for cn in xml_child_elements(parent, "param"):
            if cn.getAttribute("name") == self.name:
                use = cn.getAttribute("use")
                if use:
                    value = registry.get_object(use)
                else:
                    value_node = xml_first_element_child(cn)
                    if self.static:
                        value = self.type.value_from_xml_element(value_node, registry)
                    else:
                        value = SifAstNode.from_dom(value_node, self.type, registry)
                break
        else:
            value = self.default()

        setattr(obj, self.att_name, value)

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        param = parent.appendChild(dom.createElement("param"))
        param.setAttribute("name", self.name)
        value = getattr(obj, self.att_name)
        if isinstance(value, self._def()):
            param.setAttribute("use", ":" + value.id)
        else:
            if self.static:
                elem = self.type.value_to_xml_element(value, dom)
            else:
                elem = value.to_dom(dom, self.type)
            param.appendChild(elem)
        return param

    def from_python(self, value):
        if self.static:
            return self.type.type_wrapper(value)
        if not isinstance(value, (SifAstNode, self._def())):
            raise ValueError("%s isn't a valid value for %s" % (value, self.name))
        return value

    def default(self):
        if self.static:
            return copy.deepcopy(self.type.default_value)
        return SifValue(copy.deepcopy(self.type.default_value))


class XmlDynamicListParam(XmlDescriptor):
    _tag = "dynamic_list"

    def __init__(self, name, typename, att_name=None):
        super().__init__(name)
        self.type = TypeDescriptor(typename)
        if att_name is not None:
            self.att_name = att_name

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        param = parent.appendChild(dom.createElement("param"))
        param.setAttribute("name", self.name)
        dyl = param.appendChild(dom.createElement(self._tag))
        dyl.setAttribute("type", self.type.typename)
        values = getattr(obj, self.att_name)
        for val in values:
            entry = dyl.appendChild(dom.createElement("entry"))
            entry.appendChild(self._value_to_dom(val, dom))

    def _value_to_dom(self, val, dom: minidom.Document):
        return val.to_dom(dom, self.type)

    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        values = []

        for cn in xml_child_elements(parent, "param"):
            if cn.getAttribute("name") == self.name:
                list = xml_first_element_child(cn)
                if list.getAttribute("type") != self.type.typename:
                    raise ValueError(
                        "Wrong type for %s: got %s instead of %s" %
                        (self.name, self.type.typename, list.getAttribute("type"))
                    )
                for entry in xml_child_elements(list, "entry"):
                    values.append(self._value_from_dom(xml_first_element_child(entry), registry))
                break

        setattr(obj, self.att_name, values)

    def _value_from_dom(self, element, registry):
        return SifAstNode.from_dom(element, self.type, registry)

    def from_python(self, value):
        return value

    def default(self):
        return []


class XmlStaticListParam(XmlDynamicListParam):
    _tag = "static_list"

    def _value_to_dom(self, val, dom: minidom.Document):
        return self.type.value_to_xml_element(val, dom)

    def _value_from_dom(self, element: minidom.Element, registry: ObjectRegistry):
        return self.type.value_from_xml_element(element, registry)



import os
import re
import base64
import mimetypes
from io import BytesIO
from .base import LottieProp, PseudoBool, Index
from .layers import Layer
from .composition import Composition
from .helpers import VisualObject


## @ingroup Lottie
class Asset(VisualObject):
    _props = [
        LottieProp("id", "id", str, False),
    ]

    @classmethod
    def _load_get_class(cls, lottiedict):
        if lottiedict.get("t", None) == 3:
            return DataSource
        if "p" in lottiedict or "u" in lottiedict:
            if "w" in lottiedict:
                return Image
            return Sound
        if "layers" in lottiedict:
            return Precomp


#ingroup Lottie
class FileAsset(Asset):
    """!
    Asset referencing a file
    """
    _props = [
        LottieProp("path", "u", str, False),
        LottieProp("file_name", "p", str, False),
        LottieProp("is_embedded", "e", PseudoBool, False),
    ]

    def __init__(self):
        super().__init__()

        ## Path to the directory containing a file
        self.path = ""
        ## Filename or data url
        self.file_name = ""
        ## Whether the file is embedded
        self.is_embedded = False

    def _id_from_file(self, file):
        if not self.id:
            if isinstance(file, str):
                self.id = os.path.basename(file)
            elif hasattr(file, "name"):
                self.id = os.path.basename(file.name)
            elif hasattr(file, "filename"):
                self.id = os.path.basename(file.filename)
            else:
                self.id = "image_%s" % id(self)

    def data(self):
        """
        Returns a tuple (format, data) with the contents of the file

        `format` is a string like "png", and `data` is just raw binary data.

        If it's impossible to fetch this info, returns (None, None)
        """
        if self.is_embedded:
            m = re.match("data:[^/]+/([^;,]+);base64,(.*)", self.file_name)
            if m:
                return m.group(1), base64.b64decode(m.group(2))
            return None, None
        path = self.path + self.file_name
        if os.path.isfile(path):
            with open(path, "rb") as imgfile:
                return os.path.splitext(path)[1][1:], imgfile.read()
        return None, None


## @ingroup Lottie
class Image(FileAsset):
    """!
        External image

        @see http://docs.aenhancers.com/sources/filesource/
    """
    _props = [
        LottieProp("height", "h", float, False),
        LottieProp("width", "w", float, False),
        LottieProp("type", "t", str, False),
    ]

    @staticmethod
    def guess_mime(file):
        if isinstance(file, str):
            filename = file
        elif hasattr(file, "name"):
            filename = file.name
        else:
            return "application/octet-stream"
        return mimetypes.guess_type(filename)

    def __init__(self, id=""):
        super().__init__()

        ## Image Height
        self.height = 0
        ## Image Width
        self.width = 0
        ## Image ID
        self.id = id
        ## If "seq", marks it as part of an image sequence
        self.type = None

    def load(self, file, format=None):
        """!
        @param file     Filename, file object, or PIL.Image.Image to load
        @param format   Format to store the image data as
        """
        from PIL import Image

        if not isinstance(file, Image.Image):
            image = Image.open(file)
        else:
            image = file

        self._id_from_file(file)

        self.path = ""
        if format is None:
            format = (image.format or "png").lower()
        self.width, self.height = image.size
        output = BytesIO()
        image.save(output, format=format)
        self.file_name = "data:image/%s;base64,%s" % (
            format,
            base64.b64encode(output.getvalue()).decode("ascii")
        )
        self.is_embedded = True
        return self

    @classmethod
    def embedded(cls, image, format=None):
        """!
        Create an object from an image file
        """
        lottie_image = cls()
        return lottie_image.load(image, format)

    @classmethod
    def linked(cls, filename):
        from PIL import Image
        image = Image.open(filename)
        lottie_image = cls()
        lottie_image._id_from_file(filename)
        lottie_image.path, lottie_image.file_name = os.path.split(filename)
        lottie_image.path += "/"
        lottie_image.width = image.width
        lottie_image.height = image.height
        return lottie_image


## @ingroup Lottie
class Precomp(Asset, Composition):
    _props = [
        LottieProp("frame_rate", "fr", float, False),
    ]

    def __init__(self, id="", animation=None):
        super().__init__()
        ## Precomp ID
        self.id = id
        self.animation = animation
        if animation:
            self.animation.assets.append(self)
        self.name = None
        self.frame_rate = None

    def _on_prepare_layer(self, layer):
        if self.animation:
            self.animation.prepare_layer(layer)

    def set_timing(self, outpoint, inpoint=0, override=True):
        for layer in self.layers:
            if override or layer.in_point is None:
                layer.in_point = inpoint
            if override or layer.out_point is None:
                layer.out_point = outpoint


#ingroup Lottie
class DataSource(FileAsset):
    """!
    External data source, usually a JSON file
    """
    _props = [
        LottieProp("type", "t", int, False),
    ]
    type = 3


#ingroup Lottie
class Sound(FileAsset):
    """!
    External sound
    """



from ..parsers.baseporter import Baseporter, Loader, ExtraOption, io_progress


class ExporterLoader(Loader):
    def __init__(self):
        super().__init__(__file__, __name__, "export")

    @property
    def exporters(self):
        return self.items

    def set_options(self, parser):
        group = parser.add_argument_group("Generic output options")
        group.add_argument(
            "--pretty", "-p",
            action="store_true",
            help="Pretty print (for formats that support it)",
        )
        group.add_argument(
            "--frame",
            type=int,
            default=0,
            help="Frame to extract (for single-image formats)",
        )

        super().set_options(parser)

        return group


exporters = ExporterLoader()
exporter = exporters.decorator



#!/usr/bin/env python3
import os
import setuptools
from functools import reduce
here = os.path.dirname(os.path.abspath(__file__))


# Works fine everywhere without `encoding` except on giltab CI *shrugs*
with open(os.path.join(here, "README.md"), "r", encoding='utf8') as fh:
    long_description = fh.read()


def find_packages(root="lottie"):
    absroot = os.path.join(here, "lib", root)
    paks = [root]
    for sub in os.listdir(absroot):
        if sub == "__pycache__":
            continue
        lname = os.path.join(root, sub)
        if os.path.isdir(os.path.join(absroot, sub)):
            paks += [lname]
            paks += find_packages(lname)
    return paks


vfdir = os.path.dirname(os.path.abspath(__file__))
if not os.path.isfile(os.path.join(vfdir, ".version_full")):
    with open(os.path.join(vfdir, "version")) as vf:
        version = vf.read().strip() + "+src"
    with open(os.path.join(vfdir, ".version_full"), "w") as vf:
        vf.write(version)
else:
    with open(os.path.join(vfdir, ".version_full")) as vf:
        version = vf.read().strip()

extras_require = {
    "trace": ["glaxnimate", "pillow"],
    "images": ["pillow"],
    "PNG": ["glaxnimate"],
    "PDF": ["cairosvg"],
    "GIF": ["glaxnimate", "pillow"],
    "text": ["fonttools"],
    "video": ["opencv-python", "pillow", "numpy"],
    "emoji": ["grapheme"],
    "GUI": ["QScintilla"],
    "AnimatedSVG": ["glaxnimate"],
}
extras_require["all"] = list(reduce(lambda a, b: a | b, map(set, extras_require.values())))

setuptools.setup(
    name="lottie",
    version=version,
    author="Mattia Basaglia",
    author_email="mattia.basaglia@gmail.com",
    description="A framework to work with lottie files and telegram animated stickers (tgs)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/mattbas/python-lottie/",
    package_dir={'': 'lib'},
    license="GNU Affero General Public License v3 or later (AGPLv3+)",
    packages=find_packages(),
    scripts=[
        os.path.join("bin", "raster_palette.py"),
        os.path.join("bin", "lottie_cat.py"),
        os.path.join("bin", "tgs_check.py"),
        os.path.join("bin", "lottie_convert.py"),
        os.path.join("bin", "lottie_diff.py"),
        os.path.join("bin", "lottie_fonts.py"),
        os.path.join("bin", "lottie_printcolor.py"),
        os.path.join("bin", "lottie_diagnostic.py"),
    ],
    keywords="telegram stickers tgs lottie svg animation",
    # https://pypi.org/classifiers/
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
        "Topic :: Multimedia :: Graphics",
    ],
    zip_safe=True,
    python_requires=">=3",
    extras_require=extras_require,
    test_suite="test",
    project_urls={
        "Code": "https://gitlab.com/mattbas/python-lottie/",
        "Documentation": "https://mattbas.gitlab.io/python-lottie/index.html",
        "Chat": "https://t.me/tgs_stuff",
        "Coverage": "https://mattbas.gitlab.io/python-lottie/coverage/",
        "Downloads": "https://mattbas.gitlab.io/python-lottie/downloads.html",
    },
    data_files=[(".", [".version_full"])],
)



try:
    import glaxnimate
    has_glaxnimate = True
except ImportError:
    has_glaxnimate = False

import json
from ..nvector import NVector, Point


def convert(animation, exporter_slug):
    with glaxnimate.environment.Headless():
        document = glaxnimate.model.Document("")
        glaxnimate.io.registry.from_slug("lottie").load(document, json.dumps(animation.to_dict()).encode("utf8"))
        return glaxnimate.io.registry.from_slug(exporter_slug).save(document)


def serialize(animation, serializer_slug, time=0):
    with glaxnimate.environment.Headless():
        document = glaxnimate.model.Document("")
        glaxnimate.io.registry.from_slug("lottie").load(document, json.dumps(animation.to_dict()).encode("utf8"))
        document.current_time = time
        return glaxnimate.io.registry.serializer_from_slug(serializer_slug).serialize([document.main])


class GlaxnimateRenderer:
    def __init__(self, animation, serializer_slug, dpi):
        self.context = glaxnimate.environment.Headless()
        self.serializer_slug = serializer_slug
        self.serializer = None
        self.document = None
        self.animation = animation
        # TODO dpi

    def __enter__(self):
        self.context.__enter__()
        self.serializer = glaxnimate.io.registry.serializer_from_slug(self.serializer_slug)
        self.document = glaxnimate.model.Document("")
        glaxnimate.io.registry.from_slug("lottie").load(self.document, json.dumps(self.animation.to_dict()).encode("utf8"))
        return self

    def __exit__(self, *a, **k):
        self.context.__exit__(*a, **k)

    def serialize(self, frame, fp):
        self.document.current_time = frame
        data = self.serializer.serialize([self.document.main])
        if fp:
            fp.write(data)
        return data


def color_to_glaxnimate(color: NVector):
    return glaxnimate.utils.Color(*(color * 255).components)


def color_from_glaxnimate(color):
    return NVector(color.red / 255., color.green / 255., color.blue / 255., color.alpha / 255.)


def point_from_glaxnimate(point):
    return Point(point.x, point.y)



import json
import string
import zipfile

from .base import exporter
from ..parsers.baseporter import ExtraOption
from ..parsers.tgs import parse_tgs
from lottie import __version__
from ..objects import assets


@exporter("dotLottie Archive", ["lottie"], [
    ExtraOption("id", help="ID of the animation", default=None),
    ExtraOption("append", help="Append animation to existing archive", action="store_true"),
    ExtraOption("revision", help="File revision", type=int, default=None),
    ExtraOption("author", help="File author", default=None),
    ExtraOption("speed", help="Playback speed", type=float, default=1),
    ExtraOption("theme_color", help="Theme color", type=str, default="#ffffff"),
    ExtraOption("no_loop", help="Disable Looping", action="store_false", dest="loop"),
    ExtraOption("no_pack", help="Don't auto-pack images", action="store_false", dest="pack_images"),
], slug="dotlottie")
def export_dotlottie(animation, file, id=None, append=False, revision=None, author=None,
                     speed=1.0, theme_color="#ffffff", loop=True, pack_images=True):

    files = {}

    if append:
        with zipfile.ZipFile(file, "r") as zf:
            with zf.open("manifest.json") as manifest:
                meta = json.load(manifest)

            for name in zf.namelist():
                if name != "manifest.json":
                    files[name] = zf.read(name)
    else:
        meta = {
            "generator": "Python Lottie " + __version__,
            "version": 1.0,
            "revision": 1,
            "author": "",
            "animations": [],
            "custom": {}
        }

    if revision is not None:
        meta["revision"] = revision

    if author is not None:
        meta["author"] = author

    if id is None:
        if animation.name:
            idok = string.ascii_letters + string.digits + "_-"
            id = "".join(filter(lambda x: x in idok, animation.name.replace(" ", "_")))
        if not id:
            id = "animation_%s" % len(meta["animations"])

    meta["animations"].append({
        "id": id,
        "speed": speed,
        "themeColor": theme_color,
        "loop": loop,
    })

    if pack_images and animation.assets:
        animation = animation.clone()
        image_no = 0
        for asset in animation.assets:
            if isinstance(asset, assets.FileAsset):
                ext, data = asset.data()
                if not ext:
                    continue
                pathname = "images/"
                while True:
                    basename = "image_%s.%s" % (image_no, ext)
                    image_no += 1
                    if pathname+basename not in files:
                        break
                files[pathname+basename] = data
                asset.path = pathname
                asset.file_name = basename
                asset.is_embedded = False

    files["manifest.json"] = json.dumps(meta)
    files["animations/%s.json" % id] = json.dumps(animation.to_dict())

    with zipfile.ZipFile(file, "w") as zf:
        for name, data in files.items():
            zf.writestr(name, data)



import io
import sys
import inspect
import traceback
import contextlib

from PyQt5.QtWidgets import *


class Console(QWidget):
    def __init__(self, *a):
        super().__init__(*a)
        self.context_global = {}
        self.context_local = {}

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.lines = QPlainTextEdit()
        self.lines.setReadOnly(True)
        layout.addWidget(self.lines)

        self.input = QLineEdit()
        self.input.returnPressed.connect(self.on_input)
        layout.addWidget(self.input)

    def set_font(self, font):
        self.lines.setFont(font)
        self.input.setFont(font)

    def define(self, name, value):
        self.context_global[name] = value

    def on_input(self):
        self.eval(self.input.text())
        self.input.clear()

    def eval(self, line):
        self.lines.appendPlainText(">>> " + line)

        try:
            try:
                expression = True
                code = compile(line, "<console>", "eval")
            except SyntaxError:
                expression = False
                code = compile(line, "<console>", "single")

            sterrout = io.StringIO()
            with contextlib.redirect_stderr(sterrout):
                with contextlib.redirect_stdout(sterrout):
                    if expression:
                        value = eval(code, self.context_global, self.context_local)
                    else:
                        value = None
                        exec(code, self.context_global, self.context_local)

            stdstreams = sterrout.getvalue()
            if stdstreams:
                if stdstreams.endswith("\n"):
                    stdstreams = stdstreams[:-1]
                self.lines.appendPlainText(stdstreams)
            if value is not None:
                self.lines.appendPlainText(repr(value))
        except Exception:
            etype, value, tb = sys.exc_info()
            # Find how many frames to print to hide the current frame and above
            parent = inspect.currentframe().f_back
            i = 0
            for frame, ln in traceback.walk_tb(tb):
                i += 1
                if frame.f_back == parent:
                    i = 0

            file = io.StringIO()
            traceback.print_exception(etype, value, tb, limit=-i, file=file)
            self.lines.appendPlainText(file.getvalue())

        self.lines.verticalScrollBar().setValue(self.lines.verticalScrollBar().maximum())



from .base import importer
from ..parsers.tgs import parse_tgs


@importer("Lottie JSON / Telegram Sticker", ["json", "tgs"], slug="lottie")
def import_tgs(file, *a, **kw):
    """
    Imports a (optionally gzipped) Lottie JSON file
    """
    return parse_tgs(file, *a, **kw)


import_lottie = import_tgs



from .utils import *
from .core_nodes import XmlDescriptor, ObjectRegistry


class XmlParamSif(XmlDescriptor):
    def __init__(self, name, child_node, default_ctor=None):
        super().__init__(name)
        self.child_node = child_node
        self.default_ctor = default_ctor or child_node

    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        for cn in xml_child_elements(parent, "param"):
            if cn.getAttribute("name") == self.name:
                value = self.child_node.from_dom(xml_first_element_child(cn), registry)
                break
        else:
            value = self.default()

        setattr(obj, self.att_name, value)

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        param = parent.appendChild(dom.createElement("param"))
        param.setAttribute("name", self.name)
        param.appendChild(getattr(obj, self.att_name).to_dom(dom))
        return param

    def clean(self, value):
        if not isinstance(value, self.child_node):
            raise ValueError("%s isn't a valid value for %s" % (value, self.name))
        return value

    def default(self):
        return self.default_ctor()


class SifNodeList:
    def __init__(self, type):
        self._items = []
        self._type = type

    def __len__(self):
        return len(self._items)

    def __iter__(self):
        return iter(self._items)

    def __getitem__(self, name):
        return self._items[name]

    def __getslice__(self, i, j):
        return self._items[i:j]

    def __setitem__(self, key, value: "Layer"):
        self.validate(value)
        self._items[key] = value

    def append(self, value: "Layer"):
        self.validate(value)
        self._items.append(value)

    def __str__(self):
        return str(self._items)

    def __repr__(self):
        return "<SifNodeList %s>" % self._items

    def validate(self, value):
        if not isinstance(value, self._type):
            raise ValueError("Not a valid object: %s" % value)


class XmlSifElement(XmlDescriptor):
    def __init__(self, name, child_node, nested=True):
        super().__init__(name)
        self.child_node = child_node
        self.nested = nested

    def default(self):
        return self.child_node()

    def from_python(self, value):
        if not isinstance(value, self.child_node):
            raise ValueError("Invalid value for %s: %s" % (self.name, value))
        return value

    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        cn = xml_first_element_child(parent, self.name, allow_none=True)
        if cn:
            if self.nested:
                element = xml_first_element_child(cn)
            else:
                element = cn
            value = self.child_node.from_dom(element, registry)
        else:
            value = self.default()

        setattr(obj, self.att_name, value)

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        value = getattr(obj, self.att_name)
        if self.nested:
            node = dom.createElement(self.name)
            parent.appendChild(node)
        else:
            node = parent
        node.appendChild(value.to_dom(dom))


class XmlList(XmlDescriptor):
    def __init__(self, child_node, name=None, wrapper_tag=None, tags=None):
        super().__init__(wrapper_tag or child_node._tag)
        self.child_node = child_node
        self.att_name = self.att_name + "s" if name is None else name
        self.wrapper_tag = wrapper_tag
        if tags is None:
            self.tags = {self.name}
        else:
            self.tags = tags

    def default(self):
        return SifNodeList(self.child_node)

    def clean(self, value):
        return value

    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        values = self.default()
        for cn in xml_child_elements(parent):
            if cn.tagName in self.tags:
                value_node = cn
                if self.wrapper_tag:
                    value_node = xml_first_element_child(cn)
                values.append(self.child_node.from_dom(value_node, registry))

        setattr(obj, self.att_name, values)

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        for value in getattr(obj, self.att_name):
            value_node = value.to_dom(dom)
            if self.wrapper_tag:
                wrapper = dom.createElement(self.wrapper_tag)
                wrapper.appendChild(value_node)
                value_node = wrapper
            parent.appendChild(value_node)


class XmlWrapper(XmlDescriptor):
    def __init__(self, name, wrapped: XmlDescriptor):
        super().__init__(name)
        self.wrapped = wrapped

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        wrapper = parent.appendChild(dom.createElement(self.name))
        self.wrapped.to_xml(obj, wrapper, dom)

    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        wrapper = xml_first_element_child(parent, self.name, True)
        if wrapper:
            return self.wrapped.from_xml(obj, wrapper, registry)
        return self.default()

    def from_python(self, value):
        return self.wrapped.from_python(value)

    def initialize_object(self, dict, obj):
        return self.wrapped.initialize_object(dict, obj)

    def clean(self, value):
        return self.wrapped.clean(value)

    def default(self):
        return self.wrapped.default()


class XmlWrapperParam(XmlWrapper):
    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        wrapper = parent.appendChild(dom.createElement("param"))
        wrapper.setAttribute("name", self.name)
        self.wrapped.to_xml(obj, wrapper, dom)

    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        for wrapper in xml_child_elements(parent, "param"):
            if wrapper.getAttribute("name") == self.name:
                return self.wrapped.from_xml(obj, wrapper, registry)
        return self.default()


class XmlBoneReference(XmlDescriptor):
    def __init__(self, name):
        super().__init__(name)

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        value = getattr(obj, self.name, None)
        if not value:
            return

        node = parent.appendChild(dom.createElement(self.name))
        value_node = node.appendChild(dom.createElement("bone_valuenode"))
        value_node.setAttribute("type", value.type)
        value_node.setAttribute("guid", value.guid)

        return node

    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        node = xml_first_element_child(parent, self.name, True)
        value = None
        if node:
            value_node = xml_first_element_child(node, "bone_valuenode", True)
            if value_node:
                value = registry.get_object(value_node.getAttribute("guid"))
                if value.type != value_node.getAttribute("type"):
                    raise ValueError("Bone type %s is not %s" % (value.type, value_node.getAttribute("type")))

        setattr(obj, self.att_name, value)

    def from_python(self, value):
        return value



import math
from .base import LottieObject, LottieProp
from ..nvector import NVector


class BezierPoint:
    def __init__(self, vertex, in_tangent=None, out_tangent=None):
        self.vertex = vertex
        self.in_tangent = in_tangent or NVector(0, 0)
        self.out_tangent = out_tangent or NVector(0, 0)

    def relative(self):
        return self

    @classmethod
    def smooth(cls, point, in_tangent):
        return cls(point, in_tangent, -in_tangent)

    @classmethod
    def from_absolute(cls, point, in_tangent=None, out_tangent=None):
        if not in_tangent:
            in_tangent = point.clone()
        if not out_tangent:
            out_tangent = point.clone()
        return BezierPoint(point, in_tangent, out_tangent)


class BezierPointView:
    """
    View for bezier point
    """
    def __init__(self, bezier, index):
        self.bezier = bezier
        self.index = index

    @property
    def vertex(self):
        return self.bezier.vertices[self.index]

    @vertex.setter
    def vertex(self, point):
        self.bezier.vertices[self.index] = point

    @property
    def in_tangent(self):
        return self.bezier.in_tangents[self.index]

    @in_tangent.setter
    def in_tangent(self, point):
        self.bezier.in_tangents[self.index] = point

    @property
    def out_tangent(self):
        return self.bezier.out_tangents[self.index]

    @out_tangent.setter
    def out_tangent(self, point):
        self.bezier.out_tangents[self.index] = point

    def relative(self):
        return self


class AbsoluteBezierPointView(BezierPointView):
    @property
    def in_tangent(self):
        return self.bezier.in_tangents[self.index] + self.vertex

    @in_tangent.setter
    def in_tangent(self, point):
        self.bezier.in_tangents[self.index] = point - self.vertex

    @property
    def out_tangent(self):
        return self.bezier.out_tangents[self.index] + self.vertex

    @out_tangent.setter
    def out_tangent(self, point):
        self.bezier.out_tangents[self.index] = point - self.vertex

    def relative(self):
        return BezierPointView(self.bezier, self.index)


class BezierView:
    def __init__(self, bezier, absolute=False):
        self.bezier = bezier
        self.is_absolute = absolute

    def point(self, index):
        if self.is_absolute:
            return AbsoluteBezierPointView(self.bezier, index)
        return BezierPointView(self.bezier, index)

    def __len__(self):
        return len(self.bezier.vertices)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return [
                self.point(i)
                for i in key
            ]
        return self.point(key)

    def __iter__(self):
        for i in range(len(self)):
            yield self.point(i)

    def append(self, point):
        if isinstance(point, NVector):
            self.bezier.add_point(point.clone())
        else:
            bpt = point.relative()
            self.bezier.add_point(bpt.vertex.clone(), bpt.in_tangent.clone(), bpt.out_tangent.clone())

    @property
    def absolute(self):
        return BezierView(self.bezier, True)


## @ingroup Lottie
class Bezier(LottieObject):
    """!
    Single bezier curve
    """
    _props = [
        LottieProp("closed", "c", bool, False),
        LottieProp("in_tangents", "i", NVector, True),
        LottieProp("out_tangents", "o", NVector, True),
        LottieProp("vertices", "v", NVector, True),
    ]

    def __init__(self):
        ## Closed property of shape
        self.closed = False
        ## Cubic bezier handles for the segments before each vertex
        self.in_tangents = []
        ## Cubic bezier handles for the segments after each vertex
        self.out_tangents = []
        ## Bezier curve vertices.
        self.vertices = []
        #self.rel_tangents = rel_tangents
        ## More convent way to  access points
        self.points = BezierView(self)

    def clone(self):
        clone = Bezier()
        clone.closed = self.closed
        clone.in_tangents = [p.clone() for p in self.in_tangents]
        clone.out_tangents = [p.clone() for p in self.out_tangents]
        clone.vertices = [p.clone() for p in self.vertices]
        #clone.rel_tangents = self.rel_tangents
        return clone

    def insert_point(self, index, pos, inp=NVector(0, 0), outp=NVector(0, 0)):
        """!
        Inserts a point at the given index
        @param index    Index to insert the point at
        @param pos      Point to add
        @param inp      Tangent entering the point, as a vector relative to @p pos
        @param outp     Tangent exiting the point, as a vector relative to @p pos
        @returns @c self, for easy chaining
        """
        self.vertices.insert(index, pos)
        self.in_tangents.insert(index, inp.clone())
        self.out_tangents.insert(index, outp.clone())
        #if not self.rel_tangents:
            #self.in_tangents[-1] += pos
            #self.out_tangents[-1] += pos
        return self

    def add_point(self, pos, inp=NVector(0, 0), outp=NVector(0, 0)):
        """!
        Appends a point to the curve
        @see insert_point
        """
        self.insert_point(len(self.vertices), pos, inp, outp)
        return self

    def add_smooth_point(self, pos, inp):
        """!
        Appends a point with symmetrical tangents
        @see insert_point
        """
        self.add_point(pos, inp, -inp)
        return self

    def close(self, closed=True):
        """!
        Updates self.closed
        @returns @c self, for easy chaining
        """
        self.closed = closed
        return self

    def point_at(self, t):
        """!
        @param t    A value between 0 and 1, percentage along the length of the curve
        @returns    The point at @p t in the curve
        """
        i, t = self._index_t(t)
        points = self._bezier_points(i, True)
        return self._solve_bezier(t, points)

    def tangent_angle_at(self, t):
        i, t = self._index_t(t)
        points = self._bezier_points(i, True)

        n = len(points) - 1
        if n > 0:
            delta = sum((
                (points[i+1] - points[i]) * n * self._solve_bezier_coeff(i, n - 1, t)
                for i in range(n)
            ), NVector(0, 0))
            return math.atan2(delta.y, delta.x)

        return 0

    def _split(self, t):
        i, t = self._index_t(t)
        cub = self._bezier_points(i, True)
        split1, split2 = self._split_segment(t, cub)
        return i, split1, split2

    def _split_segment(self, t, cub):
        if len(cub) == 2:
            k = self._solve_bezier_step(t, cub)[0]
            split1 = [cub[0], NVector(0, 0), NVector(0, 0), k]
            split2 = [k, NVector(0, 0), NVector(0, 0), cub[-1]]
            return split1, split2

        if len(cub) == 3:
            quad = cub
        else:
            quad = self._solve_bezier_step(t, cub)
        lin = self._solve_bezier_step(t, quad)
        k = self._solve_bezier_step(t, lin)[0]
        split1 = [cub[0], quad[0]-cub[0], lin[0]-k, k]
        split2 = [k, lin[-1]-k, quad[-1]-cub[-1], cub[-1]]
        return split1, split2

    def split_at(self, t):
        """!
        Get two pieces out of a Bezier curve
        @param t    A value between 0 and 1, percentage along the length of the curve
        @returns Two Bezier objects that correspond to self, but split at @p t
        """
        if len(self.vertices) < 2:
            return Bezier(), self.clone()

        i, split1, split2 = self._split(t)

        seg1 = Bezier()
        seg2 = Bezier()
        for j in range(i):
            seg1.add_point(self.vertices[j].clone(), self.in_tangents[j].clone(), self.out_tangents[j].clone())
        for j in range(i+2, len(self.vertices)):
            seg2.add_point(self.vertices[j].clone(), self.in_tangents[j].clone(), self.out_tangents[j].clone())

        seg1.add_point(split1[0], self.in_tangents[i].clone(), split1[1])
        seg1.add_point(split1[3], split1[2], split2[1])

        seg2.insert_point(0, split2[0], split1[2], split2[1])
        seg2.insert_point(1, split2[3], split2[2], self.out_tangents[i+1].clone())

        return seg1, seg2

    def segment(self, t1, t2):
        """!
        Splits a Bezier in two points and returns the segment between the
        @param t1   A value between 0 and 1, percentage along the length of the curve
        @param t2   A value between 0 and 1, percentage along the length of the curve
        @returns Bezier object that correspond to the segment between @p t1 and @p t2
        """
        if self.closed and self.vertices and self.vertices[-1] != self.vertices[0]:
            copy = self.clone()
            copy.add_point(self.vertices[0])
            copy.closed = False
            return copy.segment(t1, t2)

        if t1 > 1:
            t1 = 1
        if t2 > 1:
            t2 = 1

        if t1 > t2:
            t1, t2 = t2, t1
        elif t1 == t2:
            seg = Bezier()
            p = self.point_at(t1)
            seg.add_point(p)
            seg.add_point(p)
            return seg

        seg1, seg2 = self.split_at(t1)
        t2p = (t2-t1) / (1-t1)
        seg3, seg4 = seg2.split_at(t2p)
        return seg3

    def split_self_multi(self, positions):
        """!
        Adds more points to the Bezier
        @param positions    list of percentages along the curve
        """
        if not len(positions):
            return
        t1 = positions[0]
        seg1, seg2 = self.split_at(t1)
        self.vertices = []
        self.in_tangents = []
        self.out_tangents = []

        self.vertices = seg1.vertices[:-1]
        self.in_tangents = seg1.in_tangents[:-1]
        self.out_tangents = seg1.out_tangents[:-1]

        for t2 in positions[1:]:
            t = (t2-t1) / (1-t1)
            seg1, seg2 = seg2.split_at(t)
            t1 = t
            self.vertices += seg1.vertices[:-1]
            self.in_tangents += seg1.in_tangents[:-1]
            self.out_tangents += seg1.out_tangents[:-1]

        self.vertices += seg2.vertices
        self.in_tangents += seg2.in_tangents
        self.out_tangents += seg2.out_tangents

    def split_each_segment(self):
        """!
        Adds a point in the middle of the segment between every pair of points in the Bezier
        """
        vertices = self.vertices
        in_tangents = self.in_tangents
        out_tangents = self.out_tangents

        self.vertices = []
        self.in_tangents = []
        self.out_tangents = []

        for i in range(len(vertices)-1):
            tocut = [vertices[i], out_tangents[i]+vertices[i], in_tangents[i+1]+vertices[i+1], vertices[i+1]]
            split1, split2 = self._split_segment(0.5, tocut)
            if i:
                self.out_tangents[-1] = split1[1]
            else:
                self.add_point(vertices[0], in_tangents[0], split1[1])
            self.add_point(split1[3], split1[2], split2[1])
            self.add_point(vertices[i+1], split2[2], NVector(0, 0))

    def split_self_chunks(self, n_chunks):
        """!
        Adds points the Bezier, splitting it into @p n_chunks additional chunks.
        """
        splits = [i/n_chunks for i in range(1, n_chunks)]
        return self.split_self_multi(splits)

    def _bezier_points(self, i, optimize):
        v1 = self.vertices[i].clone()
        v2 = self.vertices[i+1].clone()
        points = [v1]
        t1 = self.out_tangents[i].clone()
        if not optimize or t1.length != 0:
            points.append(t1+v1)
        t2 = self.in_tangents[i+1].clone()
        if not optimize or t1.length != 0:
            points.append(t2+v2)
        points.append(v2)
        return points

    def _solve_bezier_step(self, t, points):
        next = []
        p1 = points[0]
        for p2 in points[1:]:
            next.append(p1 * (1-t) + p2 * t)
            p1 = p2
        return next

    def _solve_bezier_coeff(self, i, n, t):
        return (
            math.factorial(n) / (math.factorial(i) * math.factorial(n - i)) # (n choose i)
            * (t ** i) * ((1 - t) ** (n-i))
        )

    def _solve_bezier(self, t, points):
        n = len(points) - 1
        if n > 0:
            return sum((
                points[i] * self._solve_bezier_coeff(i, n, t)
                for i in range(n+1)
            ), NVector(0, 0))

        #while len(points) > 1:
            #points = self._solve_bezier_step(t, points)
        return points[0]

    def _index_t(self, t):
        if t <= 0:
            return 0, 0

        if t >= 1:
            return len(self.vertices)-2, 1

        n = len(self.vertices)-1
        for i in range(n):
            if (i+1) / n > t:
                break

        return i, (t - (i/n)) * n

    def reverse(self):
        """!
        Reverses the Bezier curve
        """
        self.vertices = list(reversed(self.vertices))
        out_tangents = list(reversed(self.in_tangents))
        in_tangents = list(reversed(self.out_tangents))
        self.in_tangents = in_tangents
        self.out_tangents = out_tangents

    """def to_absolute(self):
        if self.rel_tangents:
            self.rel_tangents = False
            for i in range(len(self.vertices)):
                p = self.vertices[i]
                self.in_tangents[i] += p
                self.out_tangents[i] += p
        return self"""

    def rounded(self, round_distance):
        cloned = Bezier()
        cloned.closed = self.closed
        # value from https://spencermortensen.com/articles/bezier-circle/
        round_corner = 0.5519

        def _get_vt(closest_index):
            closer_v = self.vertices[closest_index]
            distance = (current - closer_v).length
            new_pos_perc = min(distance/2, round_distance) / distance if distance else 0
            vert = current + (closer_v - current) * new_pos_perc
            tan = - (vert - current) * round_corner
            return vert, tan

        for i, current in enumerate(self.vertices):
            if not self.closed and (i == 0 or i == len(self.points) - 1):
                cloned.points.append(self.points[i])
            else:
                vert1, out_t = _get_vt(i - 1)
                cloned.add_point(vert1, NVector(0, 0), out_t)
                vert2, in_t = _get_vt((i+1) % len(self.points))
                cloned.add_point(vert2, in_t, NVector(0, 0))

        return cloned

    def scale(self, amount):
        for vl in (self.vertices, self.in_tangents, self.out_tangents):
            for v in vl:
                v *= amount

    def lerp(self, other, t):
        if len(other.vertices) != len(self.vertices):
            if t < 1:
                return self.clone()
            return other.clone()

        bez = Bezier()
        bez.closed = self.closed

        for vlist_name in ["vertices", "in_tangents", "out_tangents"]:
            vlist = getattr(self, vlist_name)
            olist = getattr(other, vlist_name)
            out = getattr(bez, vlist_name)
            for v, o in zip(vlist, olist):
                out.append(v.lerp(o, t))

        return bez

    def rough_length(self):
        if len(self.vertices) < 2:
            return 0
        last = self.vertices[0]
        length = 0
        for v in self.vertices[1:]:
            length += (v-last).length
            last = v
        if self.closed:
            length += (last-self.vertices[0]).length
        return length



import re
from xml.etree import ElementTree

from .base import importer


_fixcrap_re = re.compile(r'(")([^"]+)(")')


def _fixcrap_item(s) -> str:
    return s.group(1) + s.group(2).replace("&", "&amp;") + s.group(3)


def _fixcrap(s: str) -> str:
    return _fixcrap_re.sub(_fixcrap_item, s)


@importer("OpenToonz", ["tnz"])
def import_opentoonz(file):
    if isinstance(file, str):
        with open(file, "r") as fobj:
            xml_data = fobj.read()
    else:
        xml_data = file.read()
        if not isinstance(xml_data, str):
            xml_data = xml_data.decode("utf-8")

    xml_data = _fixcrap(xml_data)

    xml = ElementTree.fromstring(xml_data)

    width, height = map(float, xml.find("./properties/cameras/camera/cameraRes").text.strip().split())

    print((width, height))

    import sys; sys.exit(0)




import enum
from lottie.parsers.sif.sif.core import SifNodeMeta, FrameTime
from lottie.parsers.sif.sif.enums import Smooth
from lottie.parsers.sif.xml.utils import *
from lottie.parsers.sif.xml.core_nodes import *
from lottie.parsers.sif.xml.animatable import *
from lottie.parsers.sif.xml.wrappers import *


class SifNode(metaclass=SifNodeMeta):
    def __init__(self, **kw):
        for node in self._nodes:
            node.initialize_object(kw, self)

    def __setattr__(self, name, value):
        if name in self._nodemap:
            value = self._nodemap[name].clean(value)
        return super().__setattr__(name, value)

    @staticmethod
    def static_from_dom(cls, xml: minidom.Element, registry: ObjectRegistry):
        instance = cls()
        for node in cls._nodes:
            node.from_xml(instance, xml, registry)
        return instance

    @classmethod
    def from_dom(cls, xml: minidom.Element, registry: ObjectRegistry):
        return SifNode.static_from_dom(cls, xml, registry)

    def to_dom(self, dom: minidom.Document):
        element = dom.createElement(self._tag)
        for node in self._nodes:
            node.to_xml(self, element, dom)
        return element


class AbstractTransform(SifNode):
    _nodes = [
        XmlFixedAttribute("type", "transformation")
    ]

    @classmethod
    def from_dom(cls, xml: minidom.Element, registry: ObjectRegistry):
        if xml.tagName == "bone_link":
            return SifNode.static_from_dom(BoneLinkTransform, xml, registry)

        if xml.tagName != "composite":
            raise ValueError("Invalid transform element: %s" % xml.tagName)
        return SifNode.static_from_dom(SifTransform, xml, registry)


class SifTransform(AbstractTransform):
    _tag = "composite"

    _nodes = [
        XmlAnimatable("offset", "vector", NVector(0, 0)),
        XmlAnimatable("angle", "angle", 0.),
        XmlAnimatable("skew_angle", "angle", 0.),
        XmlAnimatable("scale", "vector", NVector(1, 1)),
    ]


class BlinePoint(SifNode):
    _tag = "composite"

    _nodes = [
        XmlFixedAttribute("type", "bline_point"),
        XmlAnimatable("point", "vector", NVector(0, 0)),
        XmlAnimatable("width", "real", 1.),
        XmlAnimatable("origin", "real", .5),
        XmlAnimatable("split", "bool", False),
        XmlAnimatable("t1", "vector"),
        XmlAnimatable("t2", "vector"),
        XmlAnimatable("split_radius", "bool", True),
        XmlAnimatable("split_angle", "bool", False),
    ]


class Bline(SifNode):
    _nodes = [
        XmlAttribute("loop", bool_str, False),
        XmlFixedAttribute("type", "bline_point"),
        XmlList(BlinePoint, "points", "entry"),
    ]


class BlendMethod(enum.Enum):
    Composite = 0
    Straight = 1
    Onto = 13
    StraightOnto = 21
    Behind = 12
    Screen = 16
    Overlay = 20
    HardLight = 17
    Multiply = 6
    Divide = 7
    Add = 4
    Subtract = 5
    Difference = 18
    Lighten = 2
    Darken = 3
    Color = 8
    Hue = 9
    Saturation = 10
    Luminosity = 11
    AlphaOver = 19
    AlphaBrighten = 14
    AlphaDarken = 15
    Alpha = 23


class BlurType(enum.Enum):
    Box = 0
    FastGaussian = 1
    CrossHatch = 2
    Gaussian = 3
    Disc = 4


class WindingStyle(enum.Enum):
    NonZero = 0
    EvenOdd = 1


class Def(SifNode):
    _nodes = [
        XmlAttribute("guid", str),
        XmlAttribute("id", str),
    ]
    _subclasses = None

    @classmethod
    def from_dom(cls, xml: minidom.Element, registry: ObjectRegistry):
        actual_class = cls
        if cls == Def:
            actual_class = Def.def_types()[xml.tagName]

        obj = SifNode.static_from_dom(actual_class, xml, registry)
        if obj.id:
            registry.register_as(obj, obj.id)
        if obj.guid:
            registry.register(obj)
        return obj

    @staticmethod
    def tags():
        return list(Def.def_types().keys())

    @staticmethod
    def def_types():
        if Def._subclasses is None:
            Def._subclasses = {}
            Def._gather_def_types(Def)
        return Def._subclasses

    @staticmethod
    def _gather_def_types(cls):
        for subcls in cls.__subclasses__():
            Def._subclasses[subcls._tag] = subcls
            Def._gather_def_types(subcls)


class Duplicate(Def):
    _nodes = [
        XmlFixedAttribute("type", "real"),
        XmlAnimatable("from", "real", 1.),
        XmlAnimatable("to", "real", 1.),
        XmlAnimatable("step", "real", 1.),
    ]

    @property
    def from_(self):
        return getattr(self, "from")

    @from_.setter
    def from_(self, value):
        setattr(self, "from", value)


class ExportedValue(Def):
    def __init__(self, id, value, typename):
        self.id = id
        self.value = value
        self.type = TypeDescriptor(typename)

    def to_dom(self, dom: minidom.Document):
        element = self.value.to_dom(dom, self.type)
        element.setAttribute("id", self.id)
        return element


class Layer(SifNode):
    _types = None

    _version = "0.1"
    _layer_type = None

    _nodes = [
        XmlAttribute("type", str),
        XmlAttribute("active", bool_str, True),
        XmlAttribute("version", str),
        XmlAttribute("exclude_from_rendering", bool_str, False),
        XmlAttribute("desc", str, ""),
    ]

    def __init__(self, **kw):
        kw.setdefault("version", self._version)
        super().__init__(**kw)
        self.type = self._layer_type

    def __repr__(self):
        return "<%s.%s %r>" % (__name__, self.__class__.__name__, self.desc or self.type)

    def to_dom(self, dom: minidom.Document):
        element = dom.createElement("layer")
        for node in self._nodes:
            node.to_xml(self, element, dom)
        return element

    @classmethod
    def from_dom(cls, xml: minidom.Element, registry: ObjectRegistry):
        actual_class = cls
        if cls == Layer:
            type = xml.getAttribute("type")
            actual_class = Layer.layer_types().get(type, Layer)

        return SifNode.static_from_dom(actual_class, xml, registry)

    @staticmethod
    def layer_types():
        if Layer._types is None:
            Layer._types = {}
            Layer._gather_layer_types(Layer)
        return Layer._types

    @staticmethod
    def _gather_layer_types(cls):
        for subcls in cls.__subclasses__():
            if subcls._layer_type:
                Layer._types[subcls._layer_type] = subcls
            Layer._gather_layer_types(subcls)


class DrawableLayer(Layer):
    _nodes = [
        XmlParam("z_depth", "real", 0.),
        XmlParam("amount", "real", 1.),
        XmlParam("blend_method", "integer", BlendMethod.Composite, BlendMethod, static=True),
    ]


class GroupLayerBase(DrawableLayer):
    _nodes = [
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParamSif("transformation", AbstractTransform, SifTransform),
        XmlWrapperParam("canvas", XmlWrapper("canvas", XmlList(Layer))),

        XmlParam("time_dilation", "real", 1.),
        XmlParam("time_offset", "time", FrameTime(0, FrameTime.Unit.Frame)),
        XmlParam("children_lock", "bool", False, static=True),
        XmlParam("outline_grow", "real", 0.),
    ]

    def add_layer(self, layer: Layer):
        self.layers.append(layer)
        return layer


class FilterGroupLayer(GroupLayerBase):
    _layer_type = "filter_group"

    _nodes = [
    ]


class GroupLayer(GroupLayerBase):
    _layer_type = "group"
    _version = "0.3"

    _nodes = [
        XmlParam("z_range", "bool", False, static=True),
        XmlParam("z_range_position", "real", 0.),
        XmlParam("z_range_depth", "real", 0.),
        XmlParam("z_range_blur", "real", 0.),
    ]


class SwitchLayer(GroupLayerBase):
    _layer_type = "switch"

    _nodes = [
        XmlParam("layer_name", "string"),
        XmlParam("layer_depth", "integer", -1),
    ]


class RectangleLayer(DrawableLayer):
    _layer_type = "rectangle"

    _nodes = [
        XmlParam("color", "color", NVector(0, 0, 0, 1)),
        XmlParam("point1", "vector", NVector(0, 0)),
        XmlParam("point2", "vector", NVector(0, 0)),
        XmlParam("expand", "real", 0.),
        XmlParam("invert", "bool", False),
        XmlParam("feather_x", "real", 0.),
        XmlParam("feather_y", "real", 0.),
        XmlParam("bevel", "real", 0.),
        XmlParam("bevCircle", "bool", True),
    ]


class CircleLayer(DrawableLayer):
    _layer_type = "circle"

    _nodes = [
        XmlParam("color", "color", NVector(0, 0, 0, 1)),
        XmlParam("radius", "real", 0.),
        XmlParam("feather", "real", 0.),
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParam("invert", "bool", False),
    ]


class SimpleCircleLayer(DrawableLayer):
    _layer_type = "simple_circle"

    _nodes = [
        XmlParam("color", "color", NVector(0, 0, 0, 1)),
        XmlParam("radius", "real", 0.),
        XmlParam("center", "vector", NVector(0, 0)),
    ]


class ComplexShape(DrawableLayer):
    _nodes = [
        XmlParam("color", "color", NVector(0, 0, 0, 1)),
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParam("invert", "bool", False),
        XmlParam("antialias", "bool", True),
        XmlParam("feather", "real", 0.),
        XmlParam("blurtype", "integer", BlurType.FastGaussian, BlurType),
        XmlParam("winding_style", "integer", WindingStyle.NonZero, WindingStyle),
    ]


class StarLayer(ComplexShape):
    _layer_type = "star"

    _nodes = [
        XmlParam("radius1", "real", 0.),
        XmlParam("radius2", "real", 0.),
        XmlParam("angle", "angle", 0.),
        XmlParam("points", "integer", 5),
        XmlParam("regular_polygon", "bool", False),
    ]


class LineCap(enum.Enum):
    Rounded = 1
    Squared = 2
    Peak = 3
    Flat = 4
    InnerRounded = 5
    OffPeak = 6


class CuspStyle(enum.Enum):
    Miter = 0
    Round = 1
    Bevel = 2


class AbstractOutline(ComplexShape):
    _nodes = [
        XmlParam("width", "real", 0.1),
        XmlParam("expand", "real", 0.),
        XmlParamSif("bline", Bline),
    ]


class OutlineLayer(AbstractOutline):
    _layer_type = "outline"

    _nodes = [
        XmlParam("sharp_cusps", "bool", True),
        XmlParam("round_tip[0]", "bool", True),
        XmlParam("round_tip[1]", "bool", True),
        XmlParam("homogeneous_width", "bool", True),
    ]

    @property
    def start_tip(self):
        return LineCap.Rounded if self.round_tip_0 else LineCap.Flat

    @property
    def end_tip(self):
        return LineCap.Rounded if self.round_tip_1 else LineCap.Flat

    @property
    def cusp_type(self):
        return CuspStyle.Miter if self.sharp_cusps else CuspStyle.Round


class AdvancedOutlineLayer(AbstractOutline):
    _layer_type = "advanced_outline"

    _nodes = [
        XmlParam("start_tip", "integer", LineCap.Rounded, LineCap),
        XmlParam("end_tip", "integer", LineCap.Rounded, LineCap),
        XmlParam("cusp_type", "integer", CuspStyle.Miter, CuspStyle),
        XmlParam("smoothness", "real", 1.),
        XmlParam("homogeneous", "bool", False),
        # TODO wplist
    ]


class PolygonLayer(ComplexShape):
    _layer_type = "polygon"

    _nodes = [
        XmlDynamicListParam("vector_list", "vector", "points"),
    ]


class RegionLayer(ComplexShape):
    _layer_type = "region"

    _nodes = [
        XmlParamSif("bline", Bline),
    ]


class FontStyle(enum.Enum):
    Normal = 0
    Oblique = 1
    Italic = 2


class TextLayer(DrawableLayer):
    _layer_type = "text"

    _nodes = [
        XmlParam("text", "string"),
        XmlParam("color", "color", NVector(0, 0, 0, 1)),
        XmlParam("family", "string"),
        XmlParam("style", "integer", FontStyle.Normal, FontStyle),
        XmlParam("weight", "integer", 400),
        XmlParam("compress", "real", 1.),
        XmlParam("vcompress", "real", 1.),
        XmlParam("size", "vector", NVector(1, 1)),
        XmlParam("orient", "vector", NVector(.5, .5)),
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParam("use_kerning", "bool", False),
        XmlParam("grid_fit", "bool", False),
        XmlParam("invert", "bool", False),
    ]


class TransformDown(Layer):
    pass


class TranslateLayer(TransformDown):
    _layer_type = "translate"

    _nodes = [
        XmlParam("origin", "vector", NVector(0, 0)),
    ]


class RotateLayer(TransformDown):
    _layer_type = "rotate"

    _nodes = [
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParam("amount", "angle", 0.),
    ]


class ScaleLayer(TransformDown):
    _layer_type = "zoom"

    _nodes = [
        XmlParam("center", "vector", NVector(0, 0)),
        XmlParam("amount", "real", 0.),
    ]


class GradientLayer(DrawableLayer):
    _nodes = [
        XmlParam("gradient", "gradient", []),
        XmlParam("loop", "bool", False),
        XmlParam("zigzag", "bool", False),
    ]


class RadialGradient(GradientLayer):
    _layer_type = "radial_gradient"

    _nodes = [
        XmlParam("center", "vector", NVector(0, 0)),
        XmlParam("radius", "real", 1),
    ]


class LinearGradient(GradientLayer):
    _layer_type = "linear_gradient"

    _nodes = [
        XmlParam("p1", "vector", NVector(0, 0)),
        XmlParam("p2", "vector", NVector(0, 0)),
    ]


class ConicalLinearGradient(DrawableLayer):
    _layer_type = "conical_gradient"

    _nodes = [
        XmlParam("gradient", "gradient", []),
        XmlParam("symmetric", "bool", False),
        XmlParam("center", "vector", NVector(0, 0)),
        XmlParam("angle", "angle", 0.),
    ]


class CurveGradient(GradientLayer):
    _layer_type = "curve_gradient"

    _nodes = [
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParam("width", "real", 0.0833333358),
        XmlParamSif("bline", Bline),
        XmlParam("perpendicular", "bool", False),
        XmlParam("fast", "bool", True),
    ]


class NoiseLayer(DrawableLayer):
    _layer_type = "noise"

    _nodes = [
        XmlParam("gradient", "gradient", []),
        XmlParam("seed", "integer", 0),
        XmlParam("size", "vector", NVector(1, 1)),
        XmlParam("smooth", "integer", Smooth.Cosine, Smooth),
        XmlParam("detail", "integer", 4),
        XmlParam("speed", "integer", 0.),
        XmlParam("turbulent", "bool", False),
        XmlParam("do_alpha", "bool", False),
        XmlParam("super_sample", "bool", False),
    ]


class SpiralGradient(DrawableLayer):
    _layer_type = "spiral_gradient"

    _nodes = [
        XmlParam("gradient", "gradient", []),
        XmlParam("center", "vector", NVector(0, 0)),
        XmlParam("radius", "real", 0.5),
        XmlParam("angle", "real", 0),
        XmlParam("clockwise", "bool", False),
    ]


class BoneRoot(SifNode):
    _tag = "bone_root"

    _nodes = [
        XmlFixedAttribute("type", "bone_object"),
        XmlAttribute("guid", str)
    ]

    @classmethod
    def from_dom(cls, xml: minidom.Element, registry: ObjectRegistry):
        if xml.tagName == "bone_root":
            val = SifNode.static_from_dom(BoneRoot, xml, registry)
        else:
            val = SifNode.static_from_dom(Bone, xml, registry)
        registry.register(val)
        return val

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.guid)


class Bone(BoneRoot):
    _tag = "bone"

    _nodes = [
        XmlWrapper("name", XmlSimpleElement("string", att_name="name")),
        XmlBoneReference("parent"),
        XmlAnimatable("origin", "vector", NVector(0, 0)),
        XmlAnimatable("angle", "angle", 0.),
        XmlAnimatable("scalelx", "real", 1.),
        XmlAnimatable("width", "real", .1),
        XmlAnimatable("scalex", "real", 1.),
        XmlAnimatable("tipwidth", "real", .1),
        XmlAnimatable("bone_depth", "real", 0.),
        XmlAnimatable("length", "real", 1.),
    ]


class BoneLinkTransform(AbstractTransform):
    _tag = "bone_link"

    _nodes = [
        XmlBoneReference("bone"),
        XmlSifElement("base_value", SifTransform),
        XmlAnimatable("translate", "bool", True),
        XmlAnimatable("rotate", "bool", True),
        XmlAnimatable("skew", "bool", True),
        XmlAnimatable("scale_x", "bool", True),
        XmlAnimatable("scale_y", "bool", True),
    ]


class SkeletonLayer(Layer):
    _layer_type = "skeleton"

    _nodes = [
        XmlParam("z_depth", "real", 0.),
        XmlParam("amount", "real", 1.),
        XmlParam("name", "string"),
        XmlStaticListParam("bones", "bone_object")
    ]


class SubsamplingType(enum.Enum):
    Constant = 0
    Linear = 1
    Hyperbolic = 2


class MotionBlurLayer(Layer):
    _layer_type = "MotionBlur"

    _nodes = [
        XmlParam("aperture", "time", FrameTime(1, FrameTime.Unit.Seconds)),
        XmlParam("subsamples_factor", "real", 1.),
        XmlParam("subsampling_type", "integer", SubsamplingType.Hyperbolic, SubsamplingType),
        XmlParam("subsample_start", "real", 0.),
        XmlParam("subsample_end", "real", 1.),
    ]


class BlurLayer(DrawableLayer):
    _layer_type = "blur"

    _nodes = [
        XmlParam("size", "vector", NVector(1, 1)),
        XmlParam("type", "integer", BlurType.FastGaussian, BlurType),
    ]


class RadialBlurLayer(DrawableLayer):
    _layer_type = "radial_blur"

    _nodes = [
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParam("size", "real", .2),
        XmlParam("fade_out", "bool", False),
    ]


class CurveWarpLayer(Layer):
    _layer_type = "curve_warp"

    _nodes = [
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParam("perp_width", "real", 1.),
        XmlParam("start_point", "vector", NVector(0, 0)),
        XmlParam("end_point", "vector", NVector(0, 0)),
        XmlParamSif("bline", Bline),
        XmlParam("fast", "bool", True),
    ]


class InsideOutLayer(Layer):
    _layer_type = "inside_out"

    _nodes = [
        XmlParam("origin", "vector", NVector(0, 0)),
    ]


class NoiseDistortLayer(DrawableLayer):
    _layer_type = "noise_distort"

    _nodes = [
        XmlParam("displacement", "vector", NVector(0.25, 0.25)),
        XmlParam("size", "vector", NVector(1, 1)),
        XmlParam("seed", "integer", 0),
        XmlParam("smooth", "integer", Smooth.Cosine, Smooth),
        XmlParam("detail", "integer", 4),
        XmlParam("speed", "real", 0.),
        XmlParam("turbulent", "bool", False),
    ]


class SkeletonDeformationLayer(DrawableLayer):
    _layer_type = "skeleton_deformation"

    _nodes = [
        XmlParam("displacement", "vector", NVector(0.25, 0.25)),
        XmlParam("point1", "vector", NVector(0, 0)),
        XmlParam("point2", "vector", NVector(0, 0)),
        XmlParam("x_subdivisions", "integer", 32),
        XmlParam("y_subdivisions", "integer", 32),
        # TODO bones (pair_bone_object_bone_object)
    ]


class DistortType(enum.Enum):
    Spherize = 0
    VerticalBar = 1
    HorizontalBar = 2


class SpherizeLayer(Layer):
    _layer_type = "spherize"

    _nodes = [
        XmlParam("center", "vector", NVector(0., 0.)),
        XmlParam("radius", "real", 1.),
        XmlParam("amount", "real", 1.),
        XmlParam("clip", "bool", False),
        XmlParam("type", "integer", DistortType.Spherize, DistortType),
    ]


class StretchLayer(Layer):
    _layer_type = "stretch"

    _nodes = [
        XmlParam("amount", "vector", NVector(1., 1.)),
        XmlParam("center", "vector", NVector(0., 0.)),
    ]


class TwirlLayer(Layer):
    _layer_type = "twirl"

    _nodes = [
        XmlParam("center", "vector", NVector(0., 0.)),
        XmlParam("radius", "real", 1.),
        XmlParam("rotations", "real", 0.),
        XmlParam("distort_inside", "bool", True),
        XmlParam("distort_outside", "bool", False),
    ]


class WarpLayer(Layer):
    _layer_type = "warp"

    _nodes = [
        XmlParam("src_tl", "vector", NVector(0., 0.)),
        XmlParam("src_br", "vector", NVector(0., 0.)),
        XmlParam("dest_tl", "vector", NVector(0., 0.)),
        XmlParam("dest_tr", "vector", NVector(0., 0.)),
        XmlParam("dest_bl", "vector", NVector(0., 0.)),
        XmlParam("dest_br", "vector", NVector(0., 0.)),
        XmlParam("clip", "bool", True),
        XmlParam("interpolation", "integer", Smooth.Cubic, Smooth),
    ]


class MetaballsLayer(DrawableLayer):
    _layer_type = "metaballs"

    _nodes = [
        XmlParam("gradient", "gradient", []),
        XmlDynamicListParam("centers", "vector"),
        XmlDynamicListParam("radii", "real"),
        XmlDynamicListParam("weights", "real"),
        XmlParam("threshold", "real", 0.),
        XmlParam("threshold1", "real", 1.),
        XmlParam("positive", "bool", False),
    ]


class ClampLayer(Layer):
    _layer_type = "clamp"

    _nodes = [
        XmlParam("invert_negative", "bool", False),
        XmlParam("clamp_ceiling", "bool", False),
        XmlParam("ceiling", "real", 1.),
        XmlParam("floor", "real", 0.),
    ]


class ColorCorrectLayer(Layer):
    _layer_type = "colorcorrect"

    _nodes = [
        XmlParam("hue_adjust", "angle", 0.),
        XmlParam("brightness", "real", 0.),
        XmlParam("contrast", "real", 1.),
        XmlParam("exposure", "real", 0.),
        XmlParam("gamma", "real", 1.),
    ]


class HalftoneType(enum.Enum):
    Symmetric = 0
    LightOnDark = 2
    Diamond = 3
    Stripe = 4


class Halftone2Layer(DrawableLayer):
    _layer_type = "halftone2"

    _nodes = [
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParam("angle", "angle", 0.),
        XmlParam("size", "vector", NVector(0.25, 0.25)),
        XmlParam("color_light", "color", NVector(1, 1, 1, 1)),
        XmlParam("color_dark", "color", NVector(0, 0, 0, 1)),
        XmlParam("type", "integer", HalftoneType.Symmetric, HalftoneType),
    ]


class Halftone3Layer(DrawableLayer):
    _layer_type = "halftone3"

    _nodes = [
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParam("size", "vector", NVector(0.25, 0.25)),
        XmlParam("type", "integer", HalftoneType.Symmetric, HalftoneType),
        XmlParam("subtractive", "bool", True),

        XmlParam("color[0]", "color", NVector(0, 1, 1, 1)),
        XmlParam("tone[0].origin", "vector", NVector(0, 0)),
        XmlParam("tone[0].angle", "angle", 0.),

        XmlParam("color[1]", "color", NVector(1, 0, 1, 1)),
        XmlParam("tone[1].origin", "vector", NVector(0, 0)),
        XmlParam("tone[1].angle", "angle", 30.),

        XmlParam("color[2]", "color", NVector(1, 1, 0, 1)),
        XmlParam("tone[2].origin", "vector", NVector(0, 0)),
        XmlParam("tone[2].angle", "angle", 60.),
    ]


class LumakeyLayer(DrawableLayer):
    _layer_type = "lumakey"

    _nodes = [
    ]


class JuliaLayer(DrawableLayer):
    _layer_type = "julia"

    _nodes = [
        XmlParam("icolor", "color", NVector(0, 0, 0, 1)),
        XmlParam("ocolor", "color", NVector(0, 0, 0, 1)),
        XmlParam("color_shift", "real", 0.),
        XmlParam("iterations", "integer", 32),
        XmlParam("seed", "vector", NVector(0, 0)),
        XmlParam("bailout", "real", 2.),
        XmlParam("distort_inside", "bool", True),
        XmlParam("shade_inside", "bool", True),
        XmlParam("solid_inside", "bool", False),
        XmlParam("invert_inside", "bool", False),
        XmlParam("color_inside", "bool", True),
        XmlParam("distort_outside", "bool", True),
        XmlParam("shade_outside", "bool", True),
        XmlParam("solid_outside", "bool", False),
        XmlParam("invert_outside", "bool", False),
        XmlParam("color_outside", "bool", False),
        XmlParam("color_cycle", "bool", False),
        XmlParam("smooth_outside", "bool", True),
        XmlParam("broken", "bool", False),
    ]


class MandelbrotLayer(Layer):
    _layer_type = "mandelbrot"

    _nodes = [
        XmlParam("iterations", "integer", 32),
        XmlParam("bailout", "real", 2.),
        XmlParam("broken", "bool", False),
        XmlParam("distort_inside", "bool", True),
        XmlParam("shade_inside", "bool", True),
        XmlParam("solid_inside", "bool", False),
        XmlParam("invert_inside", "bool", False),
        XmlParam("distort_outside", "bool", True),
        XmlParam("shade_outside", "bool", True),
        XmlParam("solid_outside", "bool", False),
        XmlParam("invert_outside", "bool", False),
        XmlParam("smooth_outside", "bool", True),

        XmlParam("gradient_inside", "gradient", []),
        XmlParam("gradient_offset_inside", "real", 0.),
        XmlParam("gradient_loop_inside", "bool", True),

        XmlParam("gradient_outside", "gradient", []),
        XmlParam("gradient_offset_outside", "real", 0.),
        XmlParam("gradient_loop_outside", "bool", True),
        XmlParam("gradient_scale_outside", "real", 1.),
    ]


class CheckerboardLayer(DrawableLayer):
    _layer_type = "checker_board"

    _nodes = [
        XmlParam("color", "color", NVector(0, 0, 0, 1)),
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParam("size", "vector", NVector(0.25, 0.25)),
        XmlParam("antialias", "bool", True),
    ]


class SolidColorLayer(DrawableLayer):
    _layer_type = "SolidColor"

    _nodes = [
        XmlParam("color", "color", NVector(0, 0, 0, 1)),
    ]


class DuplicateLayer(DrawableLayer):
    _layer_type = "duplicate"

    _nodes = [
        XmlParam("index", "real"),
    ]


class ImportedImageLayer(DrawableLayer):
    _layer_type = "import"

    _nodes = [
        XmlParam("tl", "vector", NVector(0, 0)),
        XmlParam("br", "vector", NVector(0, 0)),
        XmlParam("c", "integer", Smooth.Linear, Smooth),
        XmlParam("gamma_adjust", "real", 1.),
        XmlParam("filename", "string"),
        XmlParam("time_offset", "time", FrameTime(0, FrameTime.Unit.Frame)),
    ]


class PlantLayer(DrawableLayer):
    _layer_type = "plant"

    _nodes = [
        XmlParamSif("bline", Bline),
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParam("gradient", "gradient", []),
        XmlParam("split_angle", "angle", 10.),
        XmlParam("gravity", "vector", NVector(0, -.1)),
        XmlParam("velocity", "real", .3),
        XmlParam("perp_velocity", "real", 0.),
        XmlParam("size", "real", 0.015),
        XmlParam("size_as_alpha", "bool", False),
        XmlParam("reverse", "bool", True),
        XmlParam("step", "real", 0.01),
        XmlParam("seed", "integer", 0),
        XmlParam("splits", "integer", 5),
        XmlParam("sprouts", "integer", 5),
        XmlParam("random_factor", "real", 0.2),
        XmlParam("drag", "real", 0.1),
        XmlParam("use_width", "bool", True),
    ]


class SoundLayer(Layer):
    _layer_type = "sound"

    _nodes = [
        XmlParam("z_depth", "real", 0.),
        XmlParam("filename", "string"),
        XmlParam("delay", "time", FrameTime(0, FrameTime.Unit.Seconds)),
        XmlParam("volume", "real", 1.),
    ]


class SuperSampleLayer(Layer):
    _layer_type = "super_sample"

    _nodes = [
        XmlParam("width", "integer", 2),
        XmlParam("height", "integer", 2),
        XmlParam("scanline", "bool", False),
        XmlParam("alpha_aware", "bool", True),
    ]


class XorPatternLayer(DrawableLayer):
    _layer_type = "xor_pattern"

    _nodes = [
        XmlParam("origin", "vecor", NVector(0, 0)),
        XmlParam("size", "vecor", NVector(0.25, 0.25)),
    ]


class BevelLayer(DrawableLayer):
    _layer_type = "bevel"

    _nodes = [
        XmlParam("type", "integer", BlurType.FastGaussian, BlurType),
        XmlParam("color1", "color", NVector(1, 1, 1, 1)),
        XmlParam("color2", "color", NVector(0, 0, 0, 1)),
        XmlParam("angle", "angle", 135.),
        XmlParam("depth", "real", .2),
        XmlParam("softness", "real", .1),
        XmlParam("use_luma", "bool", False),
        XmlParam("solid", "bool", False),
    ]


class ShadeLayer(DrawableLayer):
    _layer_type = "shade"

    _nodes = [
        XmlParam("type", "integer", BlurType.FastGaussian, BlurType),
        XmlParam("color", "color", NVector(1, 1, 1, 1)),
        XmlParam("origin", "vector", NVector(0, 0)),
        XmlParam("size", "vector", NVector(0.1, 0.1)),
        XmlParam("invert", "bool", False),
    ]


class FreeTimeLayer(Layer):
    _layer_type = "freetime"

    _nodes = [
        XmlParam("z_depth", "real", 0.),
        XmlParam("time", "time", FrameTime(0, FrameTime.Unit.Seconds)),
    ]


class StroboscopeLayer(Layer):
    _layer_type = "stroboscope"

    _nodes = [
        XmlParam("z_depth", "real", 0.),
        XmlParam("frequency", "real", 2.),
    ]


class TimeLoopLayer(Layer):
    _layer_type = "timeloop"

    _nodes = [
        XmlParam("z_depth", "real", 0.),
        XmlParam("link_time", "time", FrameTime(0, FrameTime.Unit.Seconds), static=True),
        XmlParam("local_time", "time", FrameTime(0, FrameTime.Unit.Seconds), static=True),
        XmlParam("duration", "time", FrameTime(0, FrameTime.Unit.Seconds), static=True),
        XmlParam("only_for_positive_duration", "bool", False, static=True),
        XmlParam("symmetrical", "bool", True, static=True),
    ]


class Keyframe(SifNode):
    _nodes = [
        XmlAttribute("active", bool_str, True),
        XmlAttribute("time", FrameTime, FrameTime(0, FrameTime.Unit.Frame)),
    ]


class Canvas(SifNode, ObjectRegistry):
    _nodes = [
        XmlAttribute("version"),
        XmlAttribute("width", float, 512),
        XmlAttribute("height", float, 512),
        XmlAttribute("xres", float, 2834.645752),
        XmlAttribute("yres", float, 2834.645752),
        XmlAttribute("gamma-r", float, 1.),
        XmlAttribute("gamma-g", float, 1.),
        XmlAttribute("gamma-b", float, 1.),
        XmlAttribute("view-box", NVector),
        XmlAttribute("antialias", bool, True),
        XmlAttribute("fps", float, 60),
        XmlAttribute("begin-time", FrameTime, FrameTime(0, FrameTime.Unit.Frame)),
        XmlAttribute("end-time", FrameTime, FrameTime(3, FrameTime.Unit.Seconds)),
        XmlAttribute("bgcolor", NVector, NVector(0, 0, 0, 0)),
        XmlSimpleElement("name"),
        XmlMeta("background_first_color", NVector, NVector(0.88, 0.88, 0.88)),
        XmlMeta("background_rendering", bool, False),
        XmlMeta("background_second_color", NVector, NVector(0.65, 0.65, 0.65)),
        XmlMeta("background_size", NVector, NVector(15, 15)),
        XmlMeta("grid_color", NVector, NVector(0.62, 0.62, 0.62)),
        XmlMeta("grid_show", bool, False),
        XmlMeta("grid_size", NVector, NVector(0.25, 0.25)),
        XmlMeta("grid_snap", bool, False),
        XmlMeta("guide_color", NVector, NVector(0.4, 0.4, 1)),
        XmlMeta("guide_show", bool, True),
        XmlMeta("guide_snap", bool, False),
        XmlMeta("jack_offset", float, 0),
        XmlMeta("onion_skin", bool, False),
        XmlMeta("onion_skin_future", int, 0),
        XmlMeta("onion_skin_past", int, 1),
        XmlList(Keyframe),
        XmlWrapper("defs", XmlList(Def, "defs", None, Def.tags())),
        XmlWrapper("bones", XmlList(BoneRoot, "bones", None, {"bone", "bone_root"})),
        XmlList(Layer),
    ]

    def __init__(self, **kw):
        SifNode.__init__(self, **kw)
        ObjectRegistry.__init__(self)

    def to_xml(self):
        dom = minidom.Document()
        dom.appendChild(self.to_dom(dom))
        return dom

    @classmethod
    def from_xml_file(cls, xml):
        if isinstance(xml, str):
            with open(xml, "r") as file:
                return cls.from_xml(minidom.parse(file))
        return cls.from_xml(minidom.parse(xml))

    @classmethod
    def from_xml_string(cls, xml):
        return cls.from_xml(minidom.parseString(xml))

    @classmethod
    def from_xml(cls, xml: minidom.Document):
        obj = cls.from_dom(xml.documentElement, None)
        xml.unlink()
        return obj

    @classmethod
    def from_dom(cls, xml: minidom.Element, registry: ObjectRegistry = None):
        instance = cls()
        for node in cls._nodes:
            node.from_xml(instance, xml, instance)
        return instance

    def time_to_frames(self, time: FrameTime):
        if time.unit == FrameTime.Unit.Frame:
            return time.value
        elif time.unit == FrameTime.Unit.Seconds:
            return time.value * self.fps

    def add_layer(self, layer: Layer):
        self.layers.append(layer)
        return layer

    def make_color(self, r, g, b, a=1):
        """
        Applies Gamma to the rgb values
        """
        return NVector(
            r ** self.gamma_r,
            g ** self.gamma_g,
            b ** self.gamma_b,
            a
        )


class Segment(SifNode):
    _nodes = [
        XmlAnimatable("p1", "vector"),
        XmlAnimatable("t1", "vector"),
        XmlAnimatable("p2", "vector"),
        XmlAnimatable("t2", "vector")
    ]


class WeightedVector(SifNode):
    _tag = "weighted_vector"

    _nodes = [
        XmlAnimatable("weight", "real", 1.),
        XmlAnimatable("value", "vector"),
    ]



import math
from .base import LottieObject, LottieProp, LottieEnum, NVector
from .properties import Value, MultiDimensional, GradientColors, ShapeProperty, Bezier, ColorValue, PositionValue
from ..utils.color import Color
from .helpers import Transform, VisualObject, BlendMode


class BoundingBox:
    """!
    Shape bounding box
    """
    def __init__(self, x1=None, y1=None, x2=None, y2=None):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2

    def include(self, x, y):
        """!
        Expands the box to include the point at x, y
        """
        if x is not None:
            if self.x1 is None or self.x1 > x:
                self.x1 = x
            if self.x2 is None or self.x2 < x:
                self.x2 = x
        if y is not None:
            if self.y1 is None or self.y1 > y:
                self.y1 = y
            if self.y2 is None or self.y2 < y:
                self.y2 = y

    def expand(self, other):
        """!
        Expands the bounding box to include another bounding box
        """
        self.include(other.x1, other.y1)
        self.include(other.x2, other.y2)

    def center(self):
        """!
        Center point of the bounding box
        """
        return NVector((self.x1 + self.x2) / 2, (self.y1 + self.y2) / 2)

    def isnull(self):
        """!
        Whether the box is default-initialized
        """
        return self.x1 is None or self.y2 is None

    def __repr__(self):
        return "<BoundingBox [%s, %s] - [%s, %s]>" % (self.x1, self.y1, self.x2, self.y2)

    @property
    def width(self):
        if self.isnull():
            return 0
        return self.x2 - self.x1

    @property
    def height(self):
        if self.isnull():
            return 0
        return self.y2 - self.y1

    def size(self):
        return NVector(self.width, self.height)


## @ingroup Lottie
class ShapeElement(VisualObject):
    """!
    Base class for all elements of ShapeLayer and Group
    """
    _props = [
        LottieProp("hidden", "hd", bool, False),
        LottieProp("type", "ty", str, False),
        LottieProp("blend_mode", "bm", BlendMode, False),
        LottieProp("property_index", "ix", int, False),
        LottieProp("css_class", "cl", str, False),
        LottieProp("layer_xml_id", "ln", str, False),

    ]
    ## %Shape type.
    type = None
    _shape_classses = None

    def __init__(self):
        super().__init__()
        ## Property index
        self.property_index = None
        ## Hide element
        self.hidden = None
        ## Blend mode
        self.blend_mode = None
        ## CSS class used by the SVG renderer
        self.css_class = None
        ## `id` attribute used by the SVG renderer
        self.layer_xml_id = None

    def bounding_box(self, time=0):
        """!
        Bounding box of the shape element at the given time
        """
        return BoundingBox()

    @classmethod
    def _load_get_class(cls, lottiedict):
        if not ShapeElement._shape_classses:
            ShapeElement._shape_classses = {}
            ShapeElement._load_sub(ShapeElement._shape_classses)
        return ShapeElement._shape_classses[lottiedict["ty"]]

    @classmethod
    def _load_sub(cls, dict):
        for sc in cls.__subclasses__():
            if sc.type and sc.type not in dict:
                dict[sc.type] = sc
            sc._load_sub(dict)

    def __str__(self):
        return self.name or super().__str__()


#ingroup Lottie
class ShapeDirection(LottieEnum):
    Undefined = 0
    ## Usually clockwise
    Normal = 1
    ## Usually counter clockwise
    Reversed = 3


## @ingroup Lottie
class Shape(ShapeElement):
    """!
    Drawable shape
    """
    _props = [
        LottieProp("direction", "d", ShapeDirection, False),
    ]

    def __init__(self):
        ShapeElement.__init__(self)
        ## After Effect's Direction. Direction how the shape is drawn. Used for trim path for example.
        self.direction = ShapeDirection.Normal

    def to_bezier(self):
        """!
        Returns a Path corresponding to this Shape
        """
        raise NotImplementedError()


## @ingroup Lottie
class Rect(Shape):
    """!
    A simple rectangle shape
    """
    _props = [
        LottieProp("position", "p", PositionValue, False),
        LottieProp("size", "s", MultiDimensional, False),
        LottieProp("rounded", "r", Value, False),
    ]
    ## %Shape type.
    type = "rc"

    def __init__(self, pos=None, size=None, rounded=0):
        Shape.__init__(self)
        ## Rect's position
        self.position = PositionValue(pos or NVector(0, 0))
        ## Rect's size
        self.size = MultiDimensional(size or NVector(0, 0))
        ## Rect's rounded corners
        self.rounded = Value(rounded)

    def bounding_box(self, time=0):
        pos = self.position.get_value(time)
        sz = self.size.get_value(time)

        return BoundingBox(
            pos[0] - sz[0]/2,
            pos[1] - sz[1]/2,
            pos[0] + sz[0]/2,
            pos[1] + sz[1]/2,
        )

    def to_bezier(self):
        """!
        Returns a Shape corresponding to this rect
        """
        shape = Path()
        kft = set()
        if self.position.animated:
            kft |= set(kf.time for kf in self.position.keyframes)
        if self.size.animated:
            kft |= set(kf.time for kf in self.size.keyframes)
        if self.rounded.animated:
            kft |= set(kf.time for kf in self.rounded.keyframes)
        if not kft:
            shape.shape.value = self._bezier_t(0)
        else:
            for time in sorted(kft):
                shape.shape.add_keyframe(time, self._bezier_t(time))
        return shape

    def _bezier_t(self, time):
        bezier = Bezier()
        bb = self.bounding_box(time)
        rounded = self.rounded.get_value(time)
        tl = NVector(bb.x1, bb.y1)
        tr = NVector(bb.x2, bb.y1)
        br = NVector(bb.x2, bb.y2)
        bl = NVector(bb.x1, bb.y2)

        if not self.rounded.animated and rounded == 0:
            bezier.add_point(tl)
            bezier.add_point(tr)
            bezier.add_point(br)
            bezier.add_point(bl)
        else:
            hh = NVector(rounded/2, 0)
            vh = NVector(0, rounded/2)
            hd = NVector(rounded, 0)
            vd = NVector(0, rounded)
            bezier.add_point(tl+vd, outp=-vh)
            bezier.add_point(tl+hd, -hh)
            bezier.add_point(tr-hd, outp=hh)
            bezier.add_point(tr+vd, -vh)
            bezier.add_point(br-vd, outp=vh)
            bezier.add_point(br-hd, hh)
            bezier.add_point(bl+hd, outp=-hh)
            bezier.add_point(bl-vd, vh)

        bezier.close()
        return bezier


## @ingroup Lottie
class StarType(LottieEnum):
    Star = 1
    Polygon = 2


## @ingroup Lottie
class Star(Shape):
    """!
    Star shape
    """
    _props = [
        LottieProp("position", "p", PositionValue, False),
        LottieProp("inner_radius", "ir", Value, False),
        LottieProp("inner_roundness", "is", Value, False),
        LottieProp("outer_radius", "or", Value, False),
        LottieProp("outer_roundness", "os", Value, False),
        LottieProp("rotation", "r", Value, False),
        LottieProp("points", "pt", Value, False),
        LottieProp("star_type", "sy", StarType, False),
    ]
    ## %Shape type.
    type = "sr"

    def __init__(self):
        Shape.__init__(self)
        ## Star's position
        self.position = PositionValue(NVector(0, 0))
        ## Star's inner radius. (Star only)
        self.inner_radius = Value()
        ## Star's inner roundness. (Star only)
        self.inner_roundness = Value()
        ## Star's outer radius.
        self.outer_radius = Value()
        ## Star's outer roundness.
        self.outer_roundness = Value()
        ## Star's rotation.
        self.rotation = Value()
        ## Star's number of points.
        self.points = Value(5)
        ## Star's type. Polygon or Star.
        self.star_type = StarType.Star

    def bounding_box(self, time=0):
        pos = self.position.get_value(time)
        r = self.outer_radius.get_value(time)

        return BoundingBox(
            pos[0] - r,
            pos[1] - r,
            pos[0] + r,
            pos[1] + r,
        )

    def to_bezier(self):
        """!
        Returns a Shape corresponding to this star
        """
        shape = Path()
        kft = set()
        if self.position.animated:
            kft |= set(kf.time for kf in self.position.keyframes)
        if self.inner_radius.animated:
            kft |= set(kf.time for kf in self.inner_radius.keyframes)
        if self.inner_roundness.animated:
            kft |= set(kf.time for kf in self.inner_roundness.keyframes)
        if self.points.animated:
            kft |= set(kf.time for kf in self.points.keyframes)
        if self.rotation.animated:
            kft |= set(kf.time for kf in self.rotation.keyframes)
        # TODO inner_roundness / outer_roundness
        if not kft:
            shape.shape.value = self._bezier_t(0)
        else:
            for time in sorted(kft):
                shape.shape.add_keyframe(time, self._bezier_t(time))
        return shape

    def _bezier_t(self, time):
        bezier = Bezier()
        pos = self.position.get_value(time)
        r1 = self.inner_radius.get_value(time)
        r2 = self.outer_radius.get_value(time)
        rot = -(self.rotation.get_value(time)) * math.pi / 180 + math.pi
        p = self.points.get_value(time)
        halfd = -math.pi / p

        for i in range(int(p)):
            main_angle = rot + i * halfd * 2
            dx = r2 * math.sin(main_angle)
            dy = r2 * math.cos(main_angle)
            bezier.add_point(NVector(pos.x + dx, pos.y + dy))

            if self.star_type == StarType.Star:
                dx = r1 * math.sin(main_angle+halfd)
                dy = r1 * math.cos(main_angle+halfd)
                bezier.add_point(NVector(pos.x + dx, pos.y + dy))

        bezier.close()
        return bezier


## @ingroup Lottie
class Ellipse(Shape):
    """!
    Ellipse shape
    """
    _props = [
        LottieProp("position", "p", PositionValue, False),
        LottieProp("size", "s", MultiDimensional, False),
    ]
    ## %Shape type.
    type = "el"

    def __init__(self, position=None, size=None):
        Shape.__init__(self)
        ## Ellipse's position
        self.position = MultiDimensional(position or NVector(0, 0))
        ## Ellipse's size
        self.size = MultiDimensional(size or NVector(0, 0))

    def bounding_box(self, time=0):
        pos = self.position.get_value(time)
        sz = self.size.get_value(time)

        return BoundingBox(
            pos[0] - sz[0]/2,
            pos[1] - sz[1]/2,
            pos[0] + sz[0]/2,
            pos[1] + sz[1]/2,
        )

    def to_bezier(self):
        """!
        Returns a Shape corresponding to this ellipse
        """
        shape = Path()
        kft = set()
        if self.position.animated:
            kft |= set(kf.time for kf in self.position.keyframes)
        if self.size.animated:
            kft |= set(kf.time for kf in self.size.keyframes)
        if not kft:
            shape.shape.value = self._bezier_t(0)
        else:
            for time in sorted(kft):
                shape.shape.add_keyframe(time, self._bezier_t(time))
        return shape

    def _bezier_t(self, time):
        from ..utils.ellipse import Ellipse as EllipseConverter

        position = self.position.get_value(time)
        radii = self.size.get_value(time) / 2
        el = EllipseConverter(position, radii, 0)
        bezier = el.to_bezier(0, math.pi*2)
        bezier.close()
        return bezier


## @ingroup Lottie
class Path(Shape):
    """!
    Animatable Bezier curve
    """
    _props = [
        LottieProp("shape", "ks", ShapeProperty, False),
        LottieProp("index", "ind", int, False),
    ]
    ## %Shape type.
    type = "sh"

    def __init__(self, bezier=None):
        Shape.__init__(self)
        ## Shape's vertices
        self.shape = ShapeProperty(bezier or Bezier())
        ## @todo Index?
        self.index = None

    def bounding_box(self, time=0):
        pos = self.shape.get_value(time)

        bb = BoundingBox()
        for v, i, o in zip(pos.vertices, pos.in_tangents, pos.out_tangents):
            bb.include(*v)
            bb.include(*(v+i))
            bb.include(*(v+o))

        return bb

    def to_bezier(self):
        return self.clone()


## @ingroup Lottie
class Group(ShapeElement):
    """!
    ShapeElement that can contain other shapes
    @note Shapes inside the same group will create "holes" in other shapes
    """
    _props = [
        LottieProp("number_of_properties", "np", float, False),
        LottieProp("shapes", "it", ShapeElement, True),
    ]
    ## %Shape type.
    type = "gr"

    def __init__(self):
        ShapeElement.__init__(self)
        ## Group number of properties. Used for expressions.
        self.number_of_properties = None
        ## Group list of items
        self.shapes = [TransformShape()]

    @property
    def transform(self):
        return self.shapes[-1]

    def bounding_box(self, time=0):
        bb = BoundingBox()
        for v in self.shapes:
            bb.expand(v.bounding_box(time))

        if not bb.isnull():
            mat = self.transform.to_matrix(time)
            points = [
                mat.apply(NVector(bb.x1, bb.y1)),
                mat.apply(NVector(bb.x1, bb.y2)),
                mat.apply(NVector(bb.x2, bb.y2)),
                mat.apply(NVector(bb.x2, bb.y1)),
            ]
            x1 = min(p.x for p in points)
            x2 = max(p.x for p in points)
            y1 = min(p.y for p in points)
            y2 = max(p.y for p in points)
            return BoundingBox(x1, y1, x2, y2)
        return bb

    def add_shape(self, shape):
        self.shapes.insert(-1, shape)
        return shape

    def insert_shape(self, index, shape):
        self.shapes.insert(index, shape)
        return shape

    @classmethod
    def load(cls, lottiedict):
        object = ShapeElement.load(lottiedict)

        shapes = []
        transform = None
        for obj in object.shapes:
            if isinstance(obj, TransformShape):
                if not transform:
                    transform = obj
            else:
                shapes.append(obj)

        object.shapes = shapes
        object.shapes.append(transform)
        return object


## @ingroup Lottie
class FillRule(LottieEnum):
    NonZero = 1
    EvenOdd = 2


## @ingroup Lottie
class Fill(ShapeElement):
    """!
    Solid fill color
    """
    _props = [
        LottieProp("opacity", "o", Value, False),
        LottieProp("color", "c", ColorValue, False),
        LottieProp("fill_rule", "r", FillRule, False),
    ]
    ## %Shape type.
    type = "fl"

    def __init__(self, color=None):
        ShapeElement.__init__(self)
        ## Fill Opacity
        self.opacity = Value(100)
        ## Fill Color
        self.color = ColorValue(color or Color(1, 1, 1))
        ## Fill rule
        self.fill_rule = None


## @ingroup Lottie
class GradientType(LottieEnum):
    Linear = 1
    Radial = 2


## @ingroup Lottie
class Gradient(LottieObject):
    _props = [
        LottieProp("start_point", "s", MultiDimensional, False),
        LottieProp("end_point", "e", MultiDimensional, False),
        LottieProp("gradient_type", "t", GradientType, False),
        LottieProp("highlight_length", "h", Value, False),
        LottieProp("highlight_angle", "a", Value, False),
        LottieProp("colors", "g", GradientColors, False),
    ]

    def __init__(self, colors=[]):
        ## Fill Opacity
        self.opacity = Value(100)
        ## Gradient Start Point
        self.start_point = MultiDimensional(NVector(0, 0))
        ## Gradient End Point
        self.end_point = MultiDimensional(NVector(0, 0))
        ## Gradient Type
        self.gradient_type = GradientType.Linear
        ## Gradient Highlight Length. Only if type is Radial
        self.highlight_length = Value()
        ## Highlight Angle. Only if type is Radial
        self.highlight_angle = Value()
        ## Gradient Colors
        self.colors = GradientColors(colors)


## @ingroup Lottie
class GradientFill(ShapeElement, Gradient):
    """!
    Gradient fill
    """
    _props = [
        LottieProp("opacity", "o", Value, False),
        LottieProp("fill_rule", "r", FillRule, False),
    ]
    ## %Shape type.
    type = "gf"

    def __init__(self, colors=[]):
        ShapeElement.__init__(self)
        Gradient.__init__(self, colors)
        ## Fill Opacity
        self.opacity = Value(100)
        ## Fill rule
        self.fill_rule = None


## @ingroup Lottie
class LineJoin(LottieEnum):
    Miter = 1
    Round = 2
    Bevel = 3


## @ingroup Lottie
class LineCap(LottieEnum):
    Butt = 1
    Round = 2
    Square = 3


## @ingroup Lottie
class StrokeDashType(LottieEnum):
    Dash = "d"
    Gap = "g"
    Offset = "o"


## @ingroup Lottie
class StrokeDash(VisualObject):
    _props = [
        LottieProp("type", "n", StrokeDashType, False),
        LottieProp("length", "v", Value, False),
    ]

    def __init__(self, length=0, type=StrokeDashType.Dash):
        super().__init__()
        self.name = type.name.lower()
        self.type = type
        self.length = Value(length)

    def __str__(self):
        return self.name or super().__str__()


## @ingroup Lottie
class BaseStroke(LottieObject):
    _props = [
        LottieProp("line_cap", "lc", LineCap, False),
        LottieProp("line_join", "lj", LineJoin, False),
        LottieProp("miter_limit", "ml", float, False),
        LottieProp("animated_miter_limit", "ml2", Value, False),
        LottieProp("opacity", "o", Value, False),
        LottieProp("width", "w", Value, False),
        LottieProp("dashes", "d", StrokeDash, True),
    ]

    def __init__(self, width=1):
        ## Stroke Line Cap
        self.line_cap = LineCap.Round
        ## Stroke Line Join
        self.line_join = LineJoin.Round
        ## Stroke Miter Limit. Only if Line Join is set to Miter.
        self.miter_limit = 0
        ## Animatable alternative to ml
        self.animated_miter_limit = None
        ## Stroke Opacity
        self.opacity = Value(100)
        ## Stroke Width
        self.width = Value(width)
        ## Dashes
        self.dashes = None


## @ingroup Lottie
class Stroke(ShapeElement, BaseStroke):
    """!
    Solid stroke
    """
    _props = [
        LottieProp("color", "c", ColorValue, False),
    ]
    ## %Shape type.
    type = "st"

    def __init__(self, color=None, width=1):
        ShapeElement.__init__(self)
        BaseStroke.__init__(self, width)
        ## Stroke Color
        self.color = ColorValue(color or Color(0, 0, 0))


## @ingroup Lottie
class GradientStroke(ShapeElement, BaseStroke, Gradient):
    """!
    Gradient stroke
    """
    ## %Shape type.
    type = "gs"

    def __init__(self, stroke_width=1):
        ShapeElement.__init__(self)
        BaseStroke.__init__(self, stroke_width)
        Gradient.__init__(self)

    def bounding_box(self, time=0):
        return BoundingBox()


## @ingroup Lottie
class TransformShape(ShapeElement, Transform):
    """!
    Group transform
    """
    ## %Shape type.
    type = "tr"

    def __init__(self):
        ShapeElement.__init__(self)
        Transform.__init__(self)
        self.anchor_point = MultiDimensional(NVector(0, 0))


## @ingroup Lottie
class Composite(LottieEnum):
    Above = 1
    Below = 2


## @ingroup Lottie
class RepeaterTransform(Transform):
    _props = [
        LottieProp("start_opacity", "so", Value, False),
        LottieProp("end_opacity", "eo", Value, False),
    ]

    def __init__(self):
        Transform.__init__(self)
        self.start_opacity = Value(100)
        self.end_opacity = Value(100)


## @ingroup Lottie
class Modifier(ShapeElement):
    pass


## @ingroup Lottie
class TrimMultipleShapes(LottieEnum):
    Individually = 1
    Simultaneously = 2


## @ingroup Lottie
## @todo Implement SIF Export
class Trim(Modifier):
    """
    Trims shapes into a segment
    """
    _props = [
        LottieProp("start", "s", Value, False),
        LottieProp("end", "e", Value, False),
        LottieProp("offset", "o", Value, False),
        LottieProp("multiple", "m", TrimMultipleShapes, False),
    ]
    ## %Shape type.
    type = "tm"

    def __init__(self):
        ShapeElement.__init__(self)
        ## Start of the segment, as a percentage
        self.start = Value(0)
        ## End of the segment, as a percentage
        self.end = Value(100)
        ## start/end offset, as an angle (0, 360)
        self.offset = Value(0)
        ## @todo?
        self.multiple = None


## @ingroup Lottie
class Repeater(Modifier):
    """
    Duplicates previous shapes in a group
    """
    _props = [
        LottieProp("copies", "c", Value, False),
        LottieProp("offset", "o", Value, False),
        LottieProp("composite", "m", Composite, False),
        LottieProp("transform", "tr", RepeaterTransform, False),
    ]
    ## %Shape type.
    type = "rp"

    def __init__(self, copies=1):
        Modifier.__init__(self)
        ## Number of Copies
        self.copies = Value(copies)
        ## Offset of Copies
        self.offset = Value()
        ## Composite of copies
        self.composite = Composite.Above
        ## Transform values for each repeater copy
        self.transform = RepeaterTransform()


## @ingroup Lottie
## @todo Implement SIF Export
class RoundedCorners(Modifier):
    """
    Rounds corners of other shapes
    """
    _props = [
        LottieProp("radius", "r", Value, False),
    ]
    ## %Shape type.
    type = "rd"

    def __init__(self, radius=0):
        Modifier.__init__(self)
        ## Rounded Corner Radius
        self.radius = Value(radius)


#ingroup Lottie
class MergeMode(LottieEnum):
    """!
    Boolean operation on shapes
    """
    Normal = 1
    Add = 2
    Subtract = 3
    Intersect = 4
    ExcludeIntersections = 5


## @ingroup Lottie
## @note marked as unsupported by lottie
class Merge(ShapeElement):
    _props = [
        LottieProp("merge_mode", "mm", MergeMode, False),
    ]
    ## %Shape type.
    type = "mm"

    def __init__(self):
        ShapeElement.__init__(self)
        ## Merge Mode
        self.merge_mode = MergeMode.Normal


## @ingroup Lottie
## @note marked as unsupported by lottie
class Twist(ShapeElement):
    _props = [
        LottieProp("angle", "a", Value, False),
        LottieProp("center", "c", MultiDimensional, False),
    ]
    ## %Shape type.
    type = "tw"

    def __init__(self):
        ShapeElement.__init__(self)
        self.angle = Value(0)
        self.center = MultiDimensional(NVector(0, 0))


## @ingroup Lottie
class PuckerBloat(ShapeElement):
    """
    Interpolates the shape with its center point and bezier tangents with the opposite direction
    """
    _props = [
        LottieProp("amount", "a", Value, False),
    ]
    ## %Shape type.
    type = "pb"

    def __init__(self):
        ShapeElement.__init__(self)
        self.amount = Value(0)


#ingroup Lottie
class ZigZag(ShapeElement):
    """!
    Changes the edges of affected shapes into a series of peaks and valleys of uniform size
    """
    _props = [
        LottieProp("shape_type", "ty", str, False),
        LottieProp("frequency", "r", Value, False),
        LottieProp("amplitude", "s", Value, False),
        LottieProp("point_type", "pt", Value, False),
    ]

    def __init__(self):
        super().__init__()

        self.shape_type = "zz"
        ## Number of ridges per segment
        self.frequency = None
        ## Distance between peaks and troughs
        self.amplitude = None
        ## Point type (1 = corner, 2 = smooth)
        self.point_type = None


#ingroup Lottie
class OffsetPath(ShapeElement):
    """!
    Interpolates the shape with its center point and bezier tangents with the opposite direction
    """
    _props = [
        LottieProp("shape_type", "ty", str, False),
        LottieProp("amount", "a", Value, False),
        LottieProp("line_join", "lj", LineJoin, False),
        LottieProp("miter_limit", "ml", Value, False),
    ]
    ## %Shape type.
    type = "op"

    def __init__(self):
        super().__init__()

        self.amount = Value(0)
        self.line_join = LineJoin.Round
        self.miter_limit = Value(0)



import warnings
from .base import LottieProp
from .helpers import VisualObject
from .properties import ColorValue, Value, GradientColors
from .shapes import GradientType


#ingroup Lottie
class LayerStyle(VisualObject):
    """!
    Style applied to a layer
    """
    _props = [
        LottieProp("type", "ty", int, False),
    ]
    type = None
    _classes = {}

    @classmethod
    def _load_get_class(cls, lottiedict):
        if not LayerStyle._classses:
            LayerStyle._classses = {
                sc.type: sc
                for sc in LayerStyle.__subclasses__()
            }
        type_id = lottiedict["ty"]
        if type_id not in LayerStyle._classses:
            warnings.warn("Unknown style type: %s" % type_id)
            return LayerStyle
        return LayerStyle._classses[type_id]


#ingroup Lottie
class BevelEmbossStyle(LayerStyle):
    type = 5
    _props = [
        LottieProp("type", "ty", int, False),
        LottieProp("bevel_style", "bs", Value, False),
        LottieProp("technique", "bt", Value, False),
        LottieProp("strength", "sr", Value, False),
        LottieProp("size", "s", Value, False),
        LottieProp("soften", "sf", Value, False),
        LottieProp("global_angle", "ga", Value, False),
        LottieProp("angle", "a", Value, False),
        LottieProp("altitude", "ll", Value, False),
        LottieProp("highlight_mode", "hm", Value, False),
        LottieProp("highlight_color", "hc", ColorValue, False),
        LottieProp("highlight_opacity", "ho", Value, False),
        LottieProp("shadow_mode", "sm", Value, False),
        LottieProp("shadow_color", "sc", ColorValue, False),
        LottieProp("shadow_opacity", "so", Value, False),
    ]

    def __init__(self):
        super().__init__()

        self.bevel_style = None
        self.technique = None
        self.strength = None
        self.size = None
        self.soften = None
        ## Use global light
        self.global_angle = None
        ## Local lighting angle
        self.angle = None
        ## Local lighting altitude
        self.altitude = None
        self.highlight_mode = None
        self.highlight_color = None
        self.highlight_opacity = None
        self.shadow_mode = None
        self.shadow_color = None
        self.shadow_opacity = None


#ingroup Lottie
class ColorOverlayStyle(LayerStyle):
    type = 7
    _props = [
        LottieProp("type", "ty", int, False),
        LottieProp("blend_mode", "bm", Value, False),
        LottieProp("color", "c", ColorValue, False),
        LottieProp("opacity", "so", Value, False),
    ]

    def __init__(self):
        super().__init__()

        self.blend_mode = None
        self.color = None
        self.opacity = None


#ingroup Lottie
class DropShadowStyle(LayerStyle):
    type = 1
    _props = [
        LottieProp("type", "ty", int, False),
        LottieProp("color", "c", ColorValue, False),
        LottieProp("opacity", "o", Value, False),
        LottieProp("angle", "a", Value, False),
        LottieProp("size", "s", Value, False),
        LottieProp("distance", "d", Value, False),
        LottieProp("choke_spread", "ch", Value, False),
        LottieProp("blend_mode", "bm", Value, False),
        LottieProp("noise", "no", Value, False),
        LottieProp("layer_conceal", "lc", Value, False),
    ]

    def __init__(self):
        super().__init__()

        self.color = None
        self.opacity = None
        ## Local light angle
        self.angle = None
        ## Blur size
        self.size = None
        self.distance = None
        self.choke_spread = None
        self.blend_mode = None
        self.noise = None
        ## Layer knowck out drop shadow
        self.layer_conceal = None


#ingroup Lottie
class GradientOverlayStyle(LayerStyle):
    type = 8
    _props = [
        LottieProp("type", "ty", int, False),
        LottieProp("blend_mode", "bm", Value, False),
        LottieProp("opacity", "o", Value, False),
        LottieProp("gradient", "gf", GradientColors, False),
        LottieProp("smoothness", "gs", Value, False),
        LottieProp("angle", "a", Value, False),
        LottieProp("gradient_type", "gt", GradientType, False),
        LottieProp("reverse", "re", Value, False),
        LottieProp("align", "al", Value, False),
        LottieProp("scale", "s", Value, False),
        LottieProp("offset", "of", Value, False),
    ]

    def __init__(self):
        super().__init__()

        self.blend_mode = None
        self.opacity = None
        self.gradient = None
        self.smoothness = None
        self.angle = None
        self.gradient_type = None
        self.reverse = None
        ## Align with layer
        self.align = None
        self.scale = None
        self.offset = None


#ingroup Lottie
class InnerGlowStyle(LayerStyle):
    type = 4
    _props = [
        LottieProp("type", "ty", int, False),
        LottieProp("color", "c", ColorValue, False),
        LottieProp("opacity", "o", Value, False),
        LottieProp("range", "r", Value, False),
        LottieProp("source", "sr", Value, False),
        LottieProp("choke_spread", "ch", Value, False),
        LottieProp("blend_mode", "bm", Value, False),
        LottieProp("noise", "no", Value, False),
        LottieProp("jitter", "j", Value, False),
    ]

    def __init__(self):
        super().__init__()

        self.color = None
        self.opacity = None
        self.range = None
        self.source = None
        self.choke_spread = None
        self.blend_mode = None
        self.noise = None
        self.jitter = None


#ingroup Lottie
class OuterGlowStyle(LayerStyle):
    type = 3
    _props = [
        LottieProp("type", "ty", int, False),
        LottieProp("color", "c", ColorValue, False),
        LottieProp("opacity", "o", Value, False),
        LottieProp("range", "r", Value, False),
        LottieProp("choke_spread", "ch", Value, False),
        LottieProp("blend_mode", "bm", Value, False),
        LottieProp("noise", "no", Value, False),
        LottieProp("jitter", "j", Value, False),
    ]

    def __init__(self):
        super().__init__()

        self.color = None
        self.opacity = None
        self.range = None
        self.choke_spread = None
        self.blend_mode = None
        self.noise = None
        self.jitter = None


#ingroup Lottie
class InnerShadowStyle(LayerStyle):
    type = 2
    _props = [
        LottieProp("type", "ty", int, False),
        LottieProp("color", "c", ColorValue, False),
        LottieProp("opacity", "o", Value, False),
        LottieProp("angle", "a", Value, False),
        LottieProp("size", "s", Value, False),
        LottieProp("distance", "d", Value, False),
        LottieProp("choke_spread", "ch", Value, False),
        LottieProp("blend_mode", "bm", Value, False),
        LottieProp("noise", "no", Value, False),
    ]

    def __init__(self):
        super().__init__()

        self.color = None
        self.opacity = None
        ## Local light angle
        self.angle = None
        ## Blur size
        self.size = None
        self.distance = None
        self.choke_spread = None
        self.blend_mode = None
        self.noise = None


#ingroup Lottie
class SatinStyle(LayerStyle):
    type = 6

    _props = [
        LottieProp("type", "ty", int, False),
        LottieProp("blend_mode", "bm", Value, False),
        LottieProp("color", "c", ColorValue, False),
        LottieProp("opacity", "o", Value, False),
        LottieProp("angle", "a", Value, False),
        LottieProp("distance", "d", Value, False),
        LottieProp("size", "s", Value, False),
        LottieProp("invert", "in", Value, False),
    ]

    def __init__(self):
        super().__init__()

        self.blend_mode = None
        self.color = None
        self.opacity = None
        self.angle = None
        self.distance = None
        self.size = None
        self.invert = None


#ingroup Lottie
class StrokeStyle(LayerStyle):
    """!
    Stroke / frame
    """
    type = 0

    _props = [
        LottieProp("type", "ty", int, False),
        LottieProp("size", "s", Value, False),
        LottieProp("color", "c", ColorValue, False),
    ]

    def __init__(self):
        super().__init__()

        self.size = None
        self.color = None



import operator
import math


def vop(op, a, b):
    return list(map(op, a, b))


class NVector():
    def __init__(self, *components):
        self.components = list(components)

    def __str__(self):
        return str(self.components)

    def __repr__(self):
        return "NVector(%s)" % ", ".join(map(str, self))

    def __len__(self):
        return len(self.components)

    def to_list(self):
        return list(self.components)

    def __add__(self, other):
        return type(self)(*vop(operator.add, self.components, other.components))

    def __sub__(self, other):
        return type(self)(*vop(operator.sub, self.components, other.components))

    def __mul__(self, scalar):
        if isinstance(scalar, NVector):
            return type(self)(*vop(operator.mul, self.components, scalar.components))
        return type(self)(*(c * scalar for c in self.components))

    def __truediv__(self, scalar):
        return type(self)(*(c / scalar for c in self.components))

    def __iadd__(self, other):
        self.components = vop(operator.add, self.components, other.components)
        return self

    def __isub__(self, other):
        self.components = vop(operator.sub, self.components, other.components)
        return self

    def __imul__(self, scalar):
        if isinstance(scalar, NVector):
            self.components = vop(operator.mul, self.components, scalar.components)
        else:
            self.components = [c * scalar for c in self.components]
        return self

    def __itruediv__(self, scalar):
        self.components = [c / scalar for c in self.components]
        return self

    def __neg__(self):
        return type(self)(*(-c for c in self.components))

    def __getitem__(self, key):
        if isinstance(key, slice):
            return NVector(*self.components[key])
        return self.components[key]

    def __setitem__(self, key, value):
        self.components[key] = value

    def __eq__(self, other):
        return self.components == other.components

    def __abs__(self):
        return type(self)(*(abs(c) for c in self.components))

    @property
    def length(self) -> float:
        return math.sqrt(sum(map(lambda x: x**2, self.components)))

    def dot(self, other):
        return sum(map(operator.mul, self.components, other.components))

    def clone(self):
        return NVector(*self.components)

    def lerp(self, other, t):
        return self * (1-t) + other * t

    @property
    def x(self):
        return self.components[0]

    @x.setter
    def x(self, v):
        self.components[0] = v

    @property
    def y(self):
        return self.components[1]

    @y.setter
    def y(self, v):
        self.components[1] = v

    @property
    def z(self):
        return self.components[2]

    @z.setter
    def z(self, v):
        self.components[2] = v

    def element_scaled(self, other):
        return type(self)(*vop(operator.mul, self.components, other.components))

    def cross(self, other):
        """
        @pre len(self) == len(other) == 3
        """
        a = self
        b = other
        return type(self)(
            a[1] * b[2] - a[2] * b[1],
            a[2] * b[0] - a[0] * b[2],
            a[0] * b[1] - a[1] * b[0],
        )

    @property
    def polar_angle(self):
        """
        @pre len(self) == 2
        """
        return math.atan2(self.y, self.x)


def Point(x, y):
    return NVector(x, y)


def Size(x, y):
    return NVector(x, y)


def Point3D(x, y, z):
    return NVector(x, y, z)


def PolarVector(length, theta):
    return NVector(length * math.cos(theta), length * math.sin(theta))



import enum
from xml.etree import ElementTree


class SvgHandler:
    ns_map = {
        "dc": "http://purl.org/dc/elements/1.1/",
        "cc": "http://creativecommons.org/ns#",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "svg": "http://www.w3.org/2000/svg",
        "sodipodi": "http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd",
        "inkscape": "http://www.inkscape.org/namespaces/inkscape",
        "xlink": "http://www.w3.org/1999/xlink",
    }

    def init_etree(self):
        for n, u in self.ns_map.items():
            ElementTree.register_namespace(n, u)

    def qualified(self, ns, name):
        return "{%s}%s" % (self.ns_map[ns], name)

    def simplified(self, name):
        for k, v in self.ns_map.items():
            name = name.replace("{%s}" % v, k+":")
        return name

    def unqualified(self, name):
        return name.split("}")[-1]

    def __init__(self):
        self.init_etree()


class NameMode(enum.Enum):
    NoName = 0
    Id = 1
    Inkscape = 2



from ..objects.base import LottieObject, ObjectVisitor
from ..objects.bezier import Bezier
from ..objects.helpers import Transform
from ..nvector import NVector


class Strip(ObjectVisitor):
    def __init__(self, float_round, remove_attributes={}):
        self.float_round = float_round
        self.remove_attributes = remove_attributes

    def round(self, fl):
        return round(fl, self.float_round)

    def nvector(self, value):
        value.components = list(map(self.round, value.components))
        return value

    def visit_property(self, object, property, value):
        if isinstance(value, Bezier):
            for l in ["vertices", "in_tangents", "out_tangents"]:
                setattr(value, l, [self.nvector(NVector(p.x, p.y)) for p in getattr(value, l)])
        elif property.lottie in self.remove_attributes:
            property.set(object, None)
        elif isinstance(value, float):
            property.set(object, round(value, 3))
        elif isinstance(value, NVector):
            self.nvector(value)


class TransformStip(Strip):
    def visit(self, object):
        if isinstance(object, Transform):
            self.transform_unset(object, "anchor_point", NVector(0, 0))
            self.transform_unset(object, "position", NVector(0, 0))
            #self.transform_unset(object, "scale", NVector(100, 100))
            self.transform_unset(object, "rotation", 0)
            #self.transform_unset(object, "opacity", 100)
            self.transform_unset(object, "skew", 0)
            self.transform_unset(object, "skew_axis", 0)

    def transform_unset(self, object, prop_name, value):
        prop = getattr(object, prop_name)
        if prop and not prop.animated and prop.value == value:
            setattr(object, prop_name, None)


heavy_strip = TransformStip(3, {"ind", "ix", "nm", "mn"})
float_strip = Strip(3)



from PIL import Image
import glaxnimate
from . import glaxnimate_helpers
import enum
from .. import objects
from ..nvector import NVector
from .pixel import _vectorizing_func


class QuantizationMode(enum.Enum):
    Nearest = 1
    Exact = 2


class PaletteAlgorithm:
    def get_colors(self, image, n_colors):
        pass


class KMeansPalette(PaletteAlgorithm):
    def __init__(self, iterations=100, match=glaxnimate.utils.quantize.MatchType.MostFrequent):
        self.iterations = iterations
        self.match = match

    def get_colors(self, image, n_colors):
        return glaxnimate.utils.quantize.k_means(image, n_colors, self.iterations, self.match)


class OctreePalette(PaletteAlgorithm):
    def get_colors(self, image, n_colors):
        return glaxnimate.utils.quantize.octree(image, n_colors)


class KModesPalette(PaletteAlgorithm):
    def get_colors(self, image, n_colors):
        return glaxnimate.utils.quantize.k_modes(image, n_colors)


class EdgeExclusionModesPalette(PaletteAlgorithm):
    def __init__(self, min_frequency=0.0005):
        self.min_frequency = min_frequency

    def get_colors(self, image, n_colors):
        return glaxnimate.utils.quantize.edge_exclusion_modes(image, n_colors, self.min_frequency)


class TraceOptions:
    def __init__(
        self,
        color_mode=QuantizationMode.Nearest,
        palette_algorithm=OctreePalette(),
        tolerance=100,
        stroke_width=1,
        smoothness=0.75,
        min_area=16
    ):
        self.trace_options = glaxnimate.utils.trace.TraceOptions()
        self.palette_algorithm = palette_algorithm
        self.color_mode = color_mode
        self.tolerance = tolerance
        self.stroke_width = stroke_width
        self.min_area = min_area
        self.smoothness = smoothness

    @property
    def smoothness(self):
        return self.trace_options.smoothness

    @smoothness.setter
    def smoothness(self, value):
        self.trace_options.smoothness = value

    @property
    def min_area(self):
        return self.trace_options.min_area

    @min_area.setter
    def min_area(self, value):
        self.trace_options.min_area = value

    def quantize(self, image, n_colors):
        """!
        Returns a list of RGB values
        """
        return self.palette_algorithm.get_colors(image, n_colors)

    def trace(self, image, codebook):
        """!
        Returns a list of tuple [color, data] where for each color in codebook
        data is a list of bezier

        You can get codebook from quantize()
        """

        if codebook is None or len(codebook) == 0:
            tracer = glaxnimate.utils.trace.Tracer(image, self.trace_options)
            tracer.set_target_alpha(128, False)
            return [glaxnimate.utils.Color(0, 0, 0), tracer.trace()]

        if self.color_mode == QuantizationMode.Nearest:
            return list(zip(codebook, glaxnimate.utils.trace.quantize_and_trace(image, self.trace_options, codebook)))

        mono_data = []
        tracer = glaxnimate.utils.trace.Tracer(image, self.trace_options)
        for color in codebook:
            tracer.set_target_color(color, self.tolerance)
            mono_data.append((color, tracer.trace()))

        return mono_data


class Vectorizer:
    def __init__(self, trace_options: TraceOptions):
        self.palette = None
        self.layers = {}
        self.trace_options = trace_options

    def _create_layer(self, animation, layer_name):
        layer = animation.add_layer(objects.ShapeLayer())
        if layer_name:
            self.layers[layer_name] = layer
            layer.name = layer_name
        return layer

    def prepare_layer(self, animation, layer_name=None):
        layer = self._create_layer(animation, layer_name)
        layer._max_verts = {}
        for color in self.palette:
            group = layer.add_shape(objects.Group())
            group.name = "color_%s" % color.name
            layer._max_verts[group.name] = 0
            fcol = glaxnimate_helpers.color_from_glaxnimate(color)
            group.add_shape(objects.Fill(NVector(*fcol)))
            if self.trace_options.stroke_width > 0:
                group.add_shape(objects.Stroke(NVector(*fcol), self.trace_options.stroke_width))
        return layer

    def raster_to_layer(self, animation, raster, layer_name=None):
        layer = self.prepare_layer(animation, layer_name)
        mono_data = self.trace_options.trace(raster, self.palette)
        for (color, beziers), group in zip(mono_data, layer.shapes):
            self.traced_to_shapes(group, beziers)
        return layer

    def traced_to_shapes(self, group, beziers):
        shapes = []
        for bezier in beziers:
            shape = group.insert_shape(0, objects.Path())
            shapes.append(shape)
            shape.shape.value = self.traced_to_bezier(bezier)
        return shapes

    def traced_to_bezier(self, path):
        bezier = objects.Bezier()
        for point in path:
            pos = glaxnimate_helpers.point_from_glaxnimate(point.pos)
            tan_in = glaxnimate_helpers.point_from_glaxnimate(point.tan_in - point.pos)
            tan_out = glaxnimate_helpers.point_from_glaxnimate(point.tan_out - point.pos)
            bezier.add_point(pos, tan_in, tan_out)
        if path.closed:
            bezier.closed = True
        return bezier


def raster_to_animation(
    filenames,
    n_colors=1,
    frame_delay=1,
    framerate=60,
    palette=[],
    trace_options=TraceOptions()
):
    vc = Vectorizer(trace_options)

    def callback(animation, raster, frame, time, duration):
        if vc.palette is None:
            if palette:
                vc.palette = [glaxnimate_helpers.color_to_glaxnimate(c) for c in palette]
            elif n_colors > 1:
                vc.palette = trace_options.quantize(raster, n_colors)
            else:
                vc.palette = [glaxnimate.utils.Color(0, 0, 0, 255)]
        layer = vc.raster_to_layer(animation, raster, "frame_%s" % frame)
        layer.in_point = time
        layer.out_point = layer.in_point + duration

    animation = _vectorizing_func(filenames, frame_delay, framerate, callback)

    return animation



import os
import subprocess
from . import objects, parsers, utils, exporters, nvector, importers
from .nvector import *
from .utils.color import Color

try:
    from .version import __version__
except ImportError:
    here = os.path.dirname(os.path.abspath(__file__))
    pipe = subprocess.Popen(
        ['git', 'describe', '--abbrev=0', '--tags'],
        cwd=here,
        stderr=subprocess.DEVNULL,
        stdout=subprocess.PIPE
    )
    out, err = pipe.communicate()
    if pipe.returncode == 0:
        __version__ = out.strip()[1:].decode("ascii") + "+git"
    else:
        vfn = os.path.join(os.path.dirname(os.path.dirname(here)), "version")
        if os.path.exists(vfn):
            with open(vfn) as vf:
                __version__ = vf.read().strip() + "+src"
        else:
            __version__ = "unknown"

try:
    version_tuple = tuple(map(int, __version__.split("+")[0].split("."))) if __version__ != "unknown" else (0, 0, 0)
except ValueError:
    version_tuple = (0, 0, 0)
    __version__ = "unknown"


__all__ = ["objects", "parsers", "utils", "exporters", "nvector", "NVector", "Point", "Color", "importers"]



from contextlib import contextmanager


@contextmanager
def open_file(file_or_name, mode="w"):
    if isinstance(file_or_name, str):
        obj = open(file_or_name, mode)
        try:
            yield obj
        finally:
            obj.close()
    else:
        yield file_or_name
